"""
Model source
"""


#[
from __future__ import annotations

import re as _re
from typing import (Self, Type, TypeAlias, Literal, Protocol, )
from collections.abc import (Iterable, )

from . import pages as _pages
from . import equations as _equations
from . import quantities as _quantities
from .parsers import preparser as _preparser
from .parsers import models as _models
from .parsers import common as _common
#]


__all__ = (
    "ModelSource",
)


QuantityInput: TypeAlias = tuple[str, str, tuple[str, ...]]
EquationInput: TypeAlias = tuple[str, tuple[str, str], tuple[str, ...]]


LOGLY_VARIABLE = (
    _quantities.QuantityKind.TRANSITION_VARIABLE
    | _quantities.QuantityKind.MEASUREMENT_VARIABLE
    | _quantities.QuantityKind.EXOGENOUS_VARIABLE
)


LOGLY_VARIABLE_OR_ANY_SHOCK = (
    LOGLY_VARIABLE
    | _quantities.QuantityKind.ANY_SHOCK
)


class SourceMixinProtocol(Protocol, ):
    """
    """
    #[

    @classmethod
    def from_source(
        klass,
        source: ModelSource,
        /,
        **kwargs,
    ) -> Self:
        ...

    #]


def from_string(
    klass: type[SourceMixinProtocol],
    source_string: str,
    /,
    context: dict | None = None,
    save_preparsed: str = "",
    **kwargs,
) -> SourceMixinProtocol:
    """Create a new object from model source string"""
    #[
    source, info = ModelSource.from_string(
        source_string, context=context, save_preparsed=save_preparsed,
    )
    return klass.from_source(source, context=context, **kwargs, )
    #]


def from_file(
    klass: type[SourceMixinProtocol],
    source_files: str | Iterable[str],
    /,
    **kwargs,
) -> SourceMixinProtocol:
    """Create a new object from source file(s)"""
    #[
    source_string = _combine_source_files(source_files, )
    return from_string(klass, source_string, **kwargs, )
    #]


class ModelSource:
    """
    """
    #[

    __slots__ = (
        "quantities",
        "dynamic_equations",
        "steady_equations",
        "log_variables",
        "all_but",
        "context",
    )

    def __init__(self, /) -> None:
        self.quantities = []
        self.dynamic_equations = []
        self.steady_equations = []
        self.log_variables = []
        self.all_but = []
        self.context = None

    @property
    def num_quantities(self, /) -> int:
        return len(self.quantities)

    @property
    def all_names(self, /) -> Iterable[str]:
        return [ qty.human for qty in self.quantities ]

    def add_parameters(self, names: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(names, _quantities.QuantityKind.PARAMETER)

    def add_exogenous_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.EXOGENOUS_VARIABLE)

    def add_transition_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.TRANSITION_VARIABLE)

    def add_anticipated_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.ANTICIPATED_SHOCK)

    def add_unanticipated_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.UNANTICIPATED_SHOCK)

    def add_transition_equations(self, equation_inputs: Iterable[EquationInput] | None) -> None:
        self._add_equations(equation_inputs, _equations.EquationKind.TRANSITION_EQUATION)

    def add_measurement_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.MEASUREMENT_VARIABLE)

    def add_measurement_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.MEASUREMENT_SHOCK)

    def add_measurement_equations(self, equation_inputs: Iterable[EquationInput] | None) -> None:
        self._add_equations(equation_inputs, _equations.EquationKind.MEASUREMENT_EQUATION)

    def add_steady_autovalues(self, equation_inputs: Iterable[EquationInput] | None) -> None:
        self._add_equations(equation_inputs, _equations.EquationKind.STEADY_AUTOVALUES, )

    def add_all_but(self, all_but_input: str | None) -> None:
        if not all_but_input:
            return
        self.all_but.append(all_but_input)

    def add_log_variables(self, log_variable_inputs: Iterable[str] | None, /, ) -> None:
        if not log_variable_inputs:
            return
        self.log_variables += log_variable_inputs

    def _add_quantities(self, quantity_inputs: Iterable[QuantityInput] | None, kind: _quantities.QuantityKind) -> None:
        """
        """
        if not quantity_inputs:
            return
        start_entry = self.quantities[-1].entry + 1 if self.quantities else 0
        self.quantities = self.quantities + [
            _quantities.Quantity(
                id=None,
                human=q[1].strip(),
                kind=kind,
                logly=None,
                description=q[0].strip(),
                entry=i,
                attributes=_conform_attributes(q[2], )
            )
            for i, q in enumerate(quantity_inputs, start=start_entry)
        ]

    def _add_equations(self, equation_inputs: Iterable[EquationInput] | None, kind: _equations.EquationKind, /) -> None:
        """
        """
        if not equation_inputs:
            return
        #
        def _create_equations(equation_inputs, kind, start_entry, human_func):
            return [
                _equations.Equation(
                    id=None,
                    human=human_func(ein),
                    kind=kind,
                    description=ein[0].strip(),
                    entry=i,
                    attributes=_conform_attributes(ein[2], ),
                )
                for i, ein in enumerate(equation_inputs, start=start_entry)
            ]
        #
        def _human_func_dynamic(ein):
            return _handle_white_spaces(ein[1][0])
        #
        def _human_func_steady(ein):
            return _handle_white_spaces(ein[1][1] if ein[1][1] else ein[1][0])
        #
        start_entry = self.dynamic_equations[-1].entry + 1 if self.dynamic_equations else 0
        self.dynamic_equations += _create_equations(equation_inputs, kind, start_entry, _human_func_dynamic)
        self.steady_equations += _create_equations(equation_inputs, kind, start_entry, _human_func_steady)

    def populate_logly(self, /) -> None:
        """
        """
        default_logly = self._resolve_default_logly()
        listed_logly = not default_logly
        for qty in self.quantities:
            if qty.kind not in LOGLY_VARIABLE:
                continue
            qty.logly = listed_logly if qty.human in self.log_variables else default_logly

    def _is_logly_consistent(self, /) -> bool:
        return all(a=="all-but" for a in self.all_but) or all(a=="" for a in self.all_but) if self.all_but else True

    def _resolve_default_logly(self, /) -> bool:
        if not self._is_logly_consistent():
            raise Exception("Inconsistent use of !all-but in !log-variables")
        return self.all_but.pop()=="all-but" if self.all_but else False

    @classmethod
    def from_lists(
        klass,
        /,
        transition_variables: Iterable[QuantityInput],
        transition_equations: Iterable[EquationInput], 
        anticipated_shocks: Iterable[QuantityInput] | None = None,
        unanticipated_shocks: Iterable[QuantityInput] | None = None,
        measurement_variables: Iterable[QuantityInput] | None = None,
        measurement_equations: Iterable[EquationInput] | None = None,
        measurement_shocks: Iterable[QuantityInput] | None = None,
        steady_autovalues: Iterable[EquationInput] | None = None,
        parameters: Iterable[QuantityInput] | None = None,
        exogenous_variables: Iterable[QuantityInput] | None = None,
        log_variables: Iterable[str] | None = None,
    ) -> Self:
        """
        """
        self = klass()
        self.add_transition_variables(transition_variables, )
        self.add_transition_equations(transition_equations, )
        self.add_anticipated_shocks(anticipated_shocks, )
        self.add_unanticipated_shocks(unanticipated_shocks, )
        self.add_measurement_variables(measurement_variables, )
        self.add_measurement_equations(measurement_equations, )
        self.add_measurement_shocks(measurement_shocks, )
        self.add_steady_autovalues(steady_autovalues, )
        self.add_parameters(parameters, )
        self.add_exogenous_variables(exogenous_variables, )
        self.add_log_variables(log_variables, )
        self.populate_logly()
        return self

    @classmethod
    def from_string(
        klass,
        source_string: str,
        /,
        context: dict | None = None,
        save_preparsed: str = "",
    ) -> tuple[Self, dict]:
        """
        """
        self = klass()
        self.context = context
        preparsed_string, preparser_info = _preparser.from_string(
            source_string, context=self.context, save_preparsed=save_preparsed, 
        )
        parsed_content = _models.from_string(preparsed_string)
        #
        self.add_transition_variables(parsed_content.get("transition-variables"))
        self.add_anticipated_shocks(parsed_content.get("anticipated-shocks"))
        self.add_unanticipated_shocks(parsed_content.get("unanticipated-shocks"))
        self.add_transition_equations(parsed_content.get("transition-equations"))
        self.add_measurement_variables(parsed_content.get("measurement-variables"))
        self.add_measurement_shocks(parsed_content.get("measurement-shocks"))
        self.add_measurement_equations(parsed_content.get("measurement-equations"))
        self.add_steady_autovalues(parsed_content.get("steady-autovalues"))
        self.add_parameters(parsed_content.get("parameters"))
        self.add_exogenous_variables(parsed_content.get("exogenous-variables"))
        self.add_log_variables(parsed_content.get("log-variables"))
        for i in parsed_content.get("all-but", []):
            self.add_all_but(i)
        self.populate_logly()
        #
        return self, preparser_info

    #]


def _handle_white_spaces(x: str, /) -> str:
    return _re.sub(r"[\s\n\r]+", "", x)


def _conform_attributes(attributes: str | Iterable[str] | None, /, ) -> set:
    """
    """
    #[
    if not attributes:
        return set()
    if isinstance(attributes, str):
        attributes = set((attributes, ))
    return set(a.strip() for a in attributes)
    #]


def _combine_source_files(
    source_files: str | Iterable[str],
    /,
    joint: str = "\n\n",
) -> str:
    """
    """
    #[
    if isinstance(source_files, str):
        source_files = [source_files]
    source_strings = []
    for f in source_files:
        with open(f, "r") as fid:
            source_strings.append(fid.read(), )
    return "\n\n".join(source_strings, )
    #]




"""
"""


#[
from __future__ import annotations

import numpy as _np

from .. import quantities as _quantities
from .. import pages as _pages
from .. import wrongdoings as _wrongdoings
from .. import has_variants as _has_variants
from ..databoxes.main import (Databox, )
#]


class Inlay:
    """
    """
    #[

    @_pages.reference(category="parameters", )
    def assign(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
················································································

==Assign model parameters==

Assigns parameters to a `Sequential` model. The method can assign parameters
from individual arguments, from a `Databox`, or from a `dict`.


### Assigning individual parameters ###

```
self.assign(
    name_one=value_one,
    name_two=value_two,
    # etc
)
```


### Assigning parameters from a `Databox` or a `dict` ###

```
self.assign(databox, )
```


### Input arguments ###


???+ input "self"
    `Sequential` model whose parameters will be assigned.


???+ input "name_one, name_two, ..."
    Names of the parameters to assign.


???+ input "value_one, value_two, ..."
    Values to assign to `name_one`, `name_two`, etc.


???+ input "databox"
    `Databox` or `dict` from which the parameters will be extracted and
    assigned. Any names in the `Databox` or `dict` that are not model
    parameters will be ignored.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        kwargs_to_assign = {}
        for a in args:
            kwargs_to_assign.update(a, )
        kwargs_to_assign.update(kwargs, )
        #
        kwargs_to_assign = {
            k: v
            for k, v in kwargs_to_assign.items()
            if k in self._invariant.parameter_names
        }
        #
        for v in self._variants:
            v.assign_from_dict_like(kwargs_to_assign, )

    @_pages.reference(category="parameters", )
    def check_missing_parameters(self, /, ) -> Databox:
        r"""
................................................................................

==Check for missing parameters==

Raises an error if any of the model parameters are missing values.

    self.check_missing_parameters()


### Input arguments ###


???+ input "self"
    `Sequential` model to check for missing parameters.


### Returns ###


Returns no value; raises an error if any parameters are missing, and prints the
list of missing parameter names.

................................................................................
        """
        parameters = self.get_parameters()
        missing = tuple(
            k for k, v in parameters.items()
            if _np.isnan(_np.array(v, dtype=float)).any()
        )
        if missing:
            message = ("Missing parameters: ", ) + missing
            raise _wrongdoings.IrisPieCritical(message, )

    @_pages.reference(category="parameters", )
    def get_parameters(
        self,
        /,
        unpack_singleton: bool = True,
    ) -> Databox:
        r"""
................................................................................

==Get model parameters==

Returns a `Databox` with the parameter values currently assigned within a
`Sequential` model.

    parameters = self.get_parameters(*, unpack_singleton=True, )


### Input arguments ###


???+ input "self"
    `Sequential` model whose parameters will be retrieved.

???+ input "unpack_singleton"
    If `True`, the method will unpack the parameters values for models with a
    single parameter variant.


### Returns ###


???+ return "parameters"
    `Databox` with the model parameters.

................................................................................
        """
        parameter_names = self._invariant.parameter_names
        parameters = {
            n: [ v.parameters[n] for v in self._variants ]
            for n in parameter_names
        }
        parameters = _has_variants.unpack_singleton_in_dict(
            parameters,
            self.is_singleton,
            unpack_singleton=unpack_singleton,
        )
        return Databox.from_dict(parameters, )

    #]




"""
Jacobians for dynamic period-by-period systems
"""


#[
from __future__ import annotations

import numpy as _np

from ..incidences import main as _incidence
from ..aldi.maps import (ArrayMap, )
from ..equations import (Equation, )

from . import base

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Any, )
    from collections.abc import (Collection, Iterable, )
#]


class _SteadyJacobian(base.Jacobian, ):
    """
    """
    #[

    _create_map = staticmethod(ArrayMap.static)
    _num_diff_columns = ...

    @staticmethod
    def _calculate_shape(
        eids: Collection[int],
        wrt_something: Collection[Any],
        num_columns_to_eval: int,
    ) -> tuple[int, int]:
        """
        """
        return len(eids), len(wrt_something),

    def _create_eid_to_wrts(
        self,
        equations: Iterable[Equation],
        all_wrt_qids: Iterable[int],
        /,
    ) -> dict[int, tuple[int, ...]]:
        """
        Create {eid: (qid, ... ), ...} where (qid, ... ) are qids wrt which the
        equation is defined no matter the time shift of the token
        """
        return {
            eqn.id: tuple(
                qid for qid in all_wrt_qids
                if _incidence.is_qid_in_tokens(eqn.incidence, qid, )
            )
            for eqn in equations
        }

    # ===== Implement AtomFactoryProtocol =====

    def create_data_index_for_token(
        self,
        token: _incidence.Token,
        /,
    ) -> tuple[int, slice]:
        """
        """
        return (token.qid, token.shift, )

    def create_diff_for_token(
        self,
        token: _incidence.Token,
        wrt_qids: dict[int, tuple[int]],
        /,
    ) -> _np.ndarray:
        """
        """
        if token is None:
            return _np.zeros((len(wrt_qids), self._num_diff_columns, ))
        try:
            index = wrt_qids.index(token.qid)
        except:
            index = None
        if index is None:
            return 0
        diff = _np.zeros((len(wrt_qids), self._num_diff_columns, ))
        diff[index, :] = self._create_diff_value_for_token(token, )
        return diff

    def _create_diff_value_for_token(
        self,
        token: _incidence.Token,
        /,
    ) -> _np.ndarray:
        """
        """
        ...

        #]


class FlatSteadyJacobian(_SteadyJacobian, ):
    """
    """
    #[

    _num_diff_columns = 1

    def eval(
        self,
        steady_array: _np.ndarray,
        column_offset: int | None,
        /,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(steady_array, column_offset, )
        return self._create_jacobian_matrix(diff_array, )

    def _create_diff_value_for_token(
        self,
        token: _incidence.Token,
        /,
    ) -> _np.ndarray:
        """
        """
        return 1

    #]


class NonflatSteadyJacobian(_SteadyJacobian, ):
    """
    """
    #[

    _num_diff_columns = 2
    nonflat_steady_shift = None

    def eval(
        self,
        steady_array: _np.ndarray,
        column_offset: int,
        /,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(steady_array, column_offset, )
        A = self._create_jacobian_matrix(diff_array[:, 0:1], )
        B = self._create_jacobian_matrix(diff_array[:, 1:2], )
        k = self.nonflat_steady_shift
        j = _np.vstack((
            _np.hstack((A, B, )),
            _np.hstack((A, B+k*A, )),
        ))
        return _np.vstack((
            _np.hstack((A, B, )),
            _np.hstack((A, B+k*A, )),
        ))

    def _create_diff_value_for_token(
        self,
        token: _incidence.Token,
        /,
    ) -> _np.ndarray:
        """
        """
        return 1, token.shift

    #]




"""
Steady equator
"""


#[
from __future__ import annotations

import numpy as _np
from typing import (Callable, Protocol, )
from collections.abc import (Iterable, )

from .. import equations as _equations
from . import plain as _plain
#]


class SteadyEquatorProtocol:
    """
    """
    #[
    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        ...
    #]


class SteadyEquator:
    """
    """
    #[

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        /,
        *,
        context: dict[str, Callable] | None = None,
    ) -> None:
        self._equator = _plain.PlainEquator(equations, context=context, )

    def eval(
        self,
        steady_array: _np.ndarray,
        column_offset: int,
        /,
    ) -> _np.ndarray:
        ...

    #]


class FlatSteadyEquator(SteadyEquator, ):
    """
    """
    #[

    def eval(
        self,
        steady_array: _np.ndarray,
        column_offset: int,
        /,
    ) -> _np.ndarray:
        """
        """
        return self._equator.eval(steady_array, column_offset, )

    #]


class NonflatSteadyEquator(SteadyEquator, ):
    """
    """
    #[

    nonflat_steady_shift = None

    def eval(
        self,
        steady_array: _np.ndarray,
        column_offset: int,
        /,
    ) -> _np.ndarray:
        """
        """
        k = self.nonflat_steady_shift
        return _np.hstack((
            self._equator.eval(steady_array, column_offset, ),
            self._equator.eval(steady_array, column_offset + k, ),
        ))
    #]





"""
Interface to X13-ARIMA-TRAMO-SEATS
"""


#[
from __future__ import annotations

from typing import (Literal, TYPE_CHECKING, )
import os as _os
import numpy as _np
import subprocess as _sp
import platform as _pf
import tempfile as _tf
import glob as _gl

from .. import executables as _executables
from .. import wrongdoings as _wrongdoings
from .. import has_variants as _has_variants
from .. import dates as _dates
from .. import pages as _pages

from . import _functionalize

if TYPE_CHECKING:
    from typing import (Any, )
    from types import (EllipsisType, )
    from ..series import (Series, )
#]


__all__ = ()


Mode = Literal["mult", "add", "pseudoadd", "logadd", ]


class Inlay:
    """
    """
    #[

    @_pages.reference(category="filtering", )
    def x13(
        self,
        *,
        span: _dates.Span | EllipsisType = ...,
        when_error: _wrongdoings.HOW = "warning",
        clean_up: bool = True,
        output: str = "sa",
        mode: Literal["mult", "add", "pseudoadd", "logadd", ] | None = None,
        #
        return_info: bool = False,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> dict[str, Any]:
        r"""
................................................................................

==X13-ARIMA-TRAMO-SEATS seasonal adjustment procedure==


### Function for creating new Series objects ###


```
new = irispie.x13(
    self,
    /,
    span=None,
    output="sa",
    mode=None,
    when_error="warning",
    clean_up=True,
    unpack_singleton=True,
    return_info=False,
)
```

```
new, info = irispie.x13(
    ...,
    return_info=True,
    ...,
)
```


### Class methods for changing existing Series objects in-place ###


```
self.x13(
    span=None,
    output="sa",
    mode=None,
    when_error="warning",
    clean_up=True,
    unpack_singleton=True,
    return_info=False,
)
```

```
info = self.x13(
    ...,
    return_info=True,
    ...,
)
```


### Input arguments ###


???+ input "self"
    A time `Series` object whose data will be run through the
    X13-ARIMA-TRAMO-SEATS procedure.

???+ input "span"
    A time span be specified as a `Span` object. If `span=None` or `span=...`,
    the time span goes from the first observed period to the last observed
    period in the input time series.

???+ input "output"
    The type of output to be returned by X13. The following options are
    available:

    | Output    | Description
    |-----------|-------------
    | `"sf"`    | Seasonal factors
    | `"sa"`    | Seasonally adjusted series
    | `"tc"`    | Trend-cycle
    | `"irr"`   | Irregular component

???+ input "mode"
    The mode to be used for the X13 run. The following options are available (see the
    [X13 documentation](https://www.census.gov/srd/www/x13as/)):

    | Mode          | Description
    |---------------|-------------
    | `None`        | Automatically selected
    | `"mult"`      | Multiplicative
    | `"add"`       | Additive
    | `"pseudoadd"` | Pseudo-additive
    | `"logadd"`    | Log-additive

    If `mode=None`, the mode is automatically selected based on the data. If the data is
    strictly positive or strictly negative, the multiplicative mode is used, otherwise
    the additive mode is used.

???+ input "when_error"
    The action to be taken when an error occurs. The following options are
    available:

    | Action      | Description
    |-------------|-------------
    | `"warning"` | Issue a warning
    | `"error"`   | Raise an error

???+ input "unpack_singleton"
    If `True`, unpack `info` into a plain dictionary for models with a
    single variant.

???+ input "return_info"
    If `True`, return a dictionary with information about the X13 run as another
    output argument.


### Returns ###


???+ returns "self"
    The `Series` object with the output data.

???+ returns "new"
    A new `Series` object with the output data.

???+ returns "info"
    (Only returned if `return_info=True` which is not the default behavior)
    Dictionary with information about the X13 run; `info` contains the
    following items:

    | Key | Description
    |-----|-------------
    | `mode` | The mode used for the X13 run
    | `log` | The log file from the X13 run
    | `out` | The output file from the X13 run
    | `err` | The error file from the X13 run
    | `success` | A boolean indicating whether the X13 run was successful


### Details ###


................................................................................
        """
        span = tuple(self._resolve_dates(span))
        base_start, base_end = span[0], span[-1]
        self.clip(base_start, base_end, )
        #
        mode, transform_function, flip_sign = _resolve_mode(self, mode, )
        settings = _create_settings(
            base_start, output,
            mode=mode,
            transform_function=transform_function,
            **kwargs,
        )
        spc = _create_spc_file(_TEMPLATE_SPC, settings, )
        #
        new_data = []
        out_info = []
        for variant_data in self.data.T:
            info_v = {
                "mode": mode,
                "transform_function": transform_function,
                "flip_sign": flip_sign,
            }
            #
            if flip_sign:
                variant_data = -variant_data
            #
            new_data_v = _x13_data(
                spc, base_start, variant_data, settings["x11_save"],
                info=info_v,
                clean_up=clean_up,
            )
            #
            if flip_sign:
                new_data_v = -new_data_v
            new_data.append(new_data_v.reshape(-1, 1, ))
            out_info.append(info_v)
            #
        if not all(i["success"] for i in out_info):
            message = "X13 failed to produce a result for at least one variant."
            _wrongdoings.raise_as(when_error, message, )
        #
        new_data = _np.hstack(new_data, )
        self._replace_start_and_values(base_start, new_data, )
        #
        if return_info:
            out_info = _has_variants.unpack_singleton(
                out_info, self.is_singleton,
                unpack_singleton=unpack_singleton,
            )
            return out_info
        else:
            return
    #]


for n in ("x13", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


def _x13_data(
    spc: str,
    base_start: _dates.Date,
    base_variant_data: _np.ndarray,
    x11_save: str,
    /,
    *,
    info: dict[str, Any],
    clean_up: bool = True,
) -> tuple[_np.ndarray, dict[str, Any]]:
    """
    """
    #[
    new_data = _np.full(base_variant_data.shape, _np.nan, dtype=float, )
    start_date, variant_data, output_slice = _remove_leading_trailing_nans(base_start, base_variant_data, )
    year, period = start_date.to_year_period()
    spc = spc.replace("$(series_start)", _X13_DATE_FORMAT_RESOLUTION[start_date.frequency].format(year=year, period=period, ), )
    spc = spc.replace("$(series_data)", _print_series_data(variant_data, ), )
    spc_file_name_without_ext = _write_spc_to_file(spc, )
    system_output = _execute(spc_file_name_without_ext, )
    raw_output_data, update_info = _collect_outputs(spc_file_name_without_ext, system_output, x11_save, )
    info.update(update_info, )
    if raw_output_data is not None:
        new_data[output_slice] = raw_output_data
    if clean_up:
        _clean_up(spc_file_name_without_ext, )
    return new_data
    #]


def _print_series_data(
    variant_data: _np.ndarray,
    /,
) -> str:
    """
    """
    return "\n".join(f"        {v:g}" for v in variant_data)


def _execute(
    spc_file_name_without_ext: str,
    /,
) -> _sp.CompletedProcess:
    """
    """
    #[
    return _sp.run(
        [_X13_EXECUTABLE_PATH, spc_file_name_without_ext, ],
        stdout=_sp.PIPE,
        check=True,
    )
    #]


def _collect_outputs(
    spc_file_name_without_ext: str,
    output: _sp.CompletedProcess,
    x11_save: str,
    /,
) -> tuple[_np.ndarray | None, dict[str, str]]:
    """
    """
    #[
    info = _read_out_info(spc_file_name_without_ext, )
    info["success"] = output.returncode == 0 and "ERROR" not in str(output.stdout)
    if info["success"]:
        out_data, info[x11_save] = _read_output_data(spc_file_name_without_ext, x11_save, info, )
    else:
        out_data = None
        info[x11_save] = None
    return out_data, info
    #]


def _read_output_data(
    spc_file_name_without_ext: str,
    x11_save: str,
    info: dict[str, str],
    /,
) -> tuple[tuple[float, ...], str]:
    """
    """
    #[
    with open(spc_file_name_without_ext + "." + x11_save, "rt", ) as fid:
        table = str(fid.read())
    out_data = tuple(float(i) for i in table.split()[5::2])
    return out_data, table
    #]


def _read_out_info(
    spc_file_name_without_ext: str,
    /,
) ->  dict[str, str]:
    """
    """
    info = dict()
    for ext in ["log", "out", "err", ]:
        with open(spc_file_name_without_ext + "." + ext, "rt", ) as fid:
            info[ext] = fid.read()
    return info


def _write_spc_to_file(
    spc: str,
    /,
) -> str:
    """
    """
    with _tf.NamedTemporaryFile(dir=".", mode="wt", suffix=".spc", delete=False, ) as fid:
        spc_file_name = fid.name
        fid.write(spc)
    return spc_file_name.removesuffix(".spc")


def _remove_leading_trailing_nans(
    base_start: _dates.Date,
    variant_data: _np.ndarray,
    /,
) -> tuple[_dates.Date, _np.ndarray, slice]:
    """
    """
    #[
    start_date = base_start.copy()
    num_rows = variant_data.shape[0]
    num_leading = 0
    num_trailing = 0
    while _np.isnan(variant_data[0]):
        num_leading += 1
        variant_data = variant_data[1:]
    while _np.isnan(variant_data[-1]):
        num_trailling += 1
        variant_data = variant_data[:-1]
    return start_date + num_leading, variant_data, slice(num_leading, num_rows-num_trailing, )
    #]


_TRANFORM_FUNCTION_DISPATCH = {
    "mult": "log",
}


def _resolve_mode(
    self: Series,
    mode: Mode | None,
    transform_function: str | None = None,
) -> tuple[str, str, bool]:
    """
    """
    #[
    flip_sign = False
    if mode is None:
        is_sign_strict = _np.all(self.data > 0) or _np.all(self.data < 0)
        mode = "mult" if is_sign_strict else "add"
    if transform_function is None:
        transform_function = _TRANFORM_FUNCTION_DISPATCH.get(mode, "none", )
    flip_sign = _np.all(self.data < 0)
    return mode, transform_function, flip_sign
    #]


def _create_settings(
    base_start: _dates.Date,
    output: str,
    /,
    *,
    mode: Mode,
    transform_function: str = "none",
) ->  dict[str, str]:
    """
    """
    #[
    settings = {
        "series_period": str(base_start.frequency.value),
        "x11_mode": str(mode),
        "x11_save": _X11_OUTPUT_RESOLUTION.get(output, output),
        "transform_function": str(transform_function),
    }
    return settings
    #]


def _create_spc_file(
    spc: str,
    settings: dict[str, str],
    /,
) -> str:
    """
    Replace placeholders with custom setting in the spc file
    """
    #[
    for k, v in settings.items():
        spc = spc.replace(f"$({k})", v)
    return spc
    #]


def _clean_up(
    spc_file_name_without_ext: str,
    /,
) -> None:
    """
    """
    #[
    for f in _gl.glob(spc_file_name_without_ext + ".*"):
        _os.remove(f)
    #]


_EXECUTABLES_PATH = _os.path.dirname(_executables.__file__)
_TEMPLATE_SPC_PATH = _os.path.join(_EXECUTABLES_PATH, "template.spc", )


with open(_TEMPLATE_SPC_PATH, "rt", ) as fid:
    _TEMPLATE_SPC = fid.read()


_X11_OUTPUT_RESOLUTION = {
    "sf": "d10",
    "sa": "d11",
    "tc": "d12",
    "irr": "d13",
}


_X13_DATE_FORMAT_RESOLUTION = {
    _dates.Frequency.MONTHLY: "{year:04d}.{period:02d}",
    _dates.Frequency.QUARTERLY: "{year:04d}.{period:1d}",
}


_X13_EXECUTABLE_FILE = {
    "Windows": "x13aswin.exe",
    "Linux": "x13asunix",
    "Darwin": "x13asmac",
}[_pf.system()]


_X13_EXECUTABLE_PATH = _os.path.join(_EXECUTABLES_PATH, _X13_EXECUTABLE_FILE, )





"""
Implement SlatableProtocol
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, )

from ..slatables import (Slatable, )
from ..series.main import (Series, )
from .. import quantities as _quantities

if TYPE_CHECKING:
    from typing import (Self, )
    from numbers import (Real, )
#]


class _Slatable(Slatable):
    """
    """
    #[

    @classmethod
    def for_simulate_and_kalman_filter(
        klass,
        model,
        output_kind: _quantities.Quantity,
        **kwargs,
    ) -> Self:
        """
        """
        #
        self = klass(**kwargs, )
        self.max_lag = model.max_lag
        self.max_lead = model.max_lead
        #
        # Databox names
        qid_to_name = model.create_qid_to_name()
        self.databox_names = tuple(
            qid_to_name[qid]
            for qid in sorted(qid_to_name)
        )
        #
        name_to_description = model.get_quantity_descriptions()
        self.descriptions = tuple(
            name_to_description[name]
            for name in self.databox_names
        )
        #
        # Databox validation - all variables must be time series
        variable_names = model.get_names(kind=_quantities.ANY_VARIABLE, )
        validator = (
            lambda x: isinstance(x, Series),
            "Input data for this variable is not a time series",
        )
        self.databox_validators = {
            name: validator
            for name in variable_names
        }
        #
        # Fallbacks and overwrites
        self.fallbacks = {}
        self.overwrites = model.get_parameters(unpack_singleton=False, )
        #
        shock_names = model.get_names(kind=_quantities.ANY_SHOCK, )
        shock_meds = {
            name: [float(0), ]*model.num_variants
            for name in shock_names
        }
        if self.shocks_from_data:
            self.fallbacks.update(shock_meds, )
        else:
            self.overwrites.update(shock_meds, )
        #
        shock_stds = model.get_stds(unpack_singleton=False, )
        if self.stds_from_data:
            self.fallbacks.update(shock_stds, )
        else:
            self.overwrites.update(shock_stds, )
        #
        self.qid_to_logly = model.create_qid_to_logly()
        self.output_names = model.get_names(kind=output_kind, )
        #
        return self

    @classmethod
    def for_multiply_stds(
        klass,
        model,
        fallbacks: dict[str, list[Real]] | None,
        **kwargs,
    ) -> None:
        """
        """
        #
        self = klass(**kwargs, )
        self.max_lag = 0
        self.max_lead = 0
        #
        # Databox names
        std_qids = _quantities.generate_qids_by_kind(
            model._invariant.quantities,
            kind=_quantities.ANY_STD,
        )
        qid_to_name = model.create_qid_to_name()
        self.databox_names = tuple(
            qid_to_name[qid]
            for qid in sorted(std_qids)
        )
        #
        self.descriptions = None
        #
        # Databox validation - all variables must be time series
        self.databox_validators = None
        #
        # Fallbacks and overwrites
        self.fallbacks = fallbacks
        self.overwrites = None
        #
        self.qid_to_logly = None
        self.output_names = self.databox_names
        #
        return self

    #]


class Inlay:
    """
    """
    #[

    def get_slatable_for_simulate(self, **kwargs, ) -> _Slatable:
        """
        """
        output_kind = (
            _quantities.ANY_VARIABLE
            | _quantities.ANY_SHOCK
        )
        slatable = _Slatable.for_simulate_and_kalman_filter(
            self,
            output_kind=output_kind,
            **kwargs,
        )
        #
        return slatable

    def get_slatable_for_kalman_filter(self, **kwargs, ) -> _Slatable:
        """
        """
        output_kind = (
            _quantities.ANY_VARIABLE
            | _quantities.ANY_SHOCK
            | _quantities.UNANTICIPATED_STD
            | _quantities.MEASUREMENT_STD
        )
        slatable = _Slatable.for_simulate_and_kalman_filter(
            self,
            output_kind=output_kind,
            **kwargs,
        )
        #
        return slatable

    def get_slatables_for_multiply_stds(self, **kwargs, ) -> tuple[_Slatable, _Slatable]:
        """
        """
        std_fallbacks = self.get_stds(unpack_singleton=False, )
        std_slatable = _Slatable.for_multiply_stds(self, fallbacks=std_fallbacks, **kwargs, )
        #
        multiplier_fallbacks = { name: 1 for name in std_fallbacks }
        multiplier_slatable = _Slatable.for_multiply_stds(self, fallbacks=multiplier_fallbacks, **kwargs, )
        #
        return std_slatable, multiplier_slatable

    #]




"""
Meta plans for dynamic simulations
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Self, Any, Protocol, NoReturn, )
from types import (EllipsisType, )
import itertools as _it
import numpy as _np
import textwrap as _tw

from ..conveniences import copies as _copies
from ..dates import (Period, )
from ..series.main import (Series, )
from .. import wrongdoings as _wrongdoings
from .. import pages as _pages
from . import _registers as _registers
from . import _pretty as _pretty
from . import _indexes as _indexes
from . import transforms as _transforms
#]


CHOOSE_TRANSFORM_CLASS = _transforms.CHOOSE_TRANSFORM_CLASS


__all__ = (
    "SimulationPlan", "PlanSimulate", "Plan",
    "CHOOSE_TRANSFORM_CLASS",
)


class SimulationPlannableProtocol(Protocol, ):
    """
    """
    #[

    simulate_can_be_exogenized: Iterable[str] | None
    simulate_can_be_endogenized: Iterable[str] | None
    simulate_can_be_exogenized_anticipated: Iterable[str] | None
    simulate_can_be_exogenized_anticipated: Iterable[str] | None
    simulate_can_be_endogenized_unanticipated: Iterable[str] | None
    simulate_can_be_endogenized_unanticipated: Iterable[str] | None

    #]


#[
@_pages.reference(
    path=("structural_models", "simulation_plans.md", ),
    categories={
        "constructor": "Creating new simulation plans",
        "definition_simultaneous": "Defining exogenized and endogenized data points in [`Simultaneous` simulations](simultaneous.md#simulate)",
        "definition_sequential": "Defining exogenized and endogenized data points in [`Sequential` simulations](sequential.md#simulate)",
        "information": "Getting information about simulation plans",
        "information_simultaneous": "Getting information about simulation plans for [`Simultaneous` models](simultaneous.md)",
        "information_sequential": "Getting information about simulation plans for [`Sequential` models](sequential.md)",
    },
)
#]
class SimulationPlan(
    _registers.Mixin,
    _pretty.Mixin,
    _indexes.ItemMixin,
    _copies.Mixin,
):
    """
················································································

Simulation meta plans
======================

`SimulationPlan` objects are used to set up meta information about
conditioning assumptions for simulations of
[`Simultaneous`](simultaneous_modelsd) or [`Sequential`](sequential_models)
models. The simulation plans specify

* what variables to exogenize in what periods
* what shocks to endogenized in what periods (`Simultaneous` models only)
* what anticipation status to assign (`Simultaneous` models only)

The plans only contain meta information, not the actual data points for the
exogenized variables. The actual data points are expected to be included in
the input databox when the simulation is run.

················································································
    """
    #[

    _TABLE_FIELDS = ("NAME", "PERIOD(S)", "REGISTER", "TRANSFORM", "VALUE", )

    _registers = (
        "exogenized",
        "endogenized",
        "exogenized_anticipated",
        "exogenized_unanticipated",
        "endogenized_unanticipated",
        "endogenized_anticipated",
    )

    __slots__ = (
        ("base_span", )
        + tuple(f"_can_be_{r}" for r in _registers)
        + tuple(f"_{r}_register" for r in _registers)
        + tuple(f"default_{r}" for r in _registers)
    )

    @_pages.reference(
        category="constructor",
        call_name="SimulationPlan",
    )
    def __init__(
        self,
        model,
        span: Iterable[Period] | None,
    ) -> None:
        """
················································································

==Create new simulation plan object==

```
self = SimulationPlan(model, time_span, )
```

Create a new simulation plan object for a
[`Simultaneous`](sequential_models) or
[`Sequential`](sequential_models) model.

### Input arguments ###

???+ input "model"

    A [`Simultaneous`](sequential_models) or
    [`Sequential`](sequential_models) model that will be simulated.

???+ input "time_span"

    A date range on which the `model` will be simulated.


### Returns ###

???+ returns "self"

    A new empty simulation plan object.

················································································
        """
        self.base_span = tuple(span)
        self._default_exogenized = None
        self._default_endogenized = None
        plannable = model.get_simulation_plannable()
        def default_value(*args, **kwargs, ):
            return [None] * self.num_periods
        self._initialize_registers(plannable, default_value, )

    def check_consistency(
        self,
        plannable: SimulationPlannableProtocol,
        span: Iterable[Period] | None,
        /,
    ) -> None:
        """
        """
        benchmark = type(self)(plannable, span, )
        if self.base_span != benchmark.base_span:
            raise _wrongdoings.IrisPieError(f"Plan span must be the same as the simulation span")
        for r in self._registers:
            if getattr(self, f"can_be_{r}") != getattr(benchmark, f"can_be_{r}"):
                raise _wrongdoings.IrisPieError(f"Plan must be created using the simulated model")

    @property
    @_pages.reference(category="property", )
    def start(self, /, ) -> Period:
        """==Start date of the simulation span=="""
        return self.base_span[0]

    start_period = start
    start_date = start

    @property
    @_pages.reference(category="property", )
    def end(self, /, ) -> Period:
        """==End date of the simulation span=="""
        return self.base_span[-1]

    end_period = end
    end_date = end

    @property
    @_pages.reference(category="property", )
    def num_periods(self, /, ) -> int:
        """==Number of periods in the simulation span=="""
        return len(self.base_span) if self.base_span is not None else 1

    @property
    @_pages.reference(category="property", )
    def frequency(self, /, ) -> str:
        """==Date frequency of the simulation span=="""
        return self.start.frequency

    @_pages.reference(category="definition_sequential", )
    def exogenize(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        /,
        *,
        transform: str | None = None,
        # when_data: bool | None = None,
        **kwargs,
    ) -> None:
        r"""
................................................................................

==Exogenize certain LHS quantities at certain dates==

Exogenize certain LHS quantities at specified dates, setting them as
predetermined values within the simulation of
a [`Sequential` model](sequential.md). This method is used to control how
the model behaves during simulations by fixing certain variables to known
values.

    self.exogenize(
        dates,
        names,
        *,
        transform=None,
        when_data=False,
    )

### Input arguments ###

???+ input "self"
    The simulation plan in which data points will be exogenized.

???+ input "dates"
    A list of dates or `...` to apply to all dates at which the quantities 
    will be exogenized.

???+ input "names"
    A list of names or a single name, or `...` to apply to all names that 
    specifies which quantities to set as predetermined at the specified dates.

???+ input "transform"
    Specifies the transformation to apply to the exogenized quantities. If not
    specified, no transformation is applied. Available transformations include:

    * `None`: Exogenize the LHS variables as they are with no
    transformation.

    * `"log"`: Exogenize the natural logarithm of the LHS variables. Input
    time series needs to be prefixed with `log_`.

    * `"diff"`: Exogenize the first difference of the LHS variables. Input
    time series needs to be prefixed with `diff_`.

    * `"diff_log"`: Exogenize the first difference of the natural logarithm
    of the LHS variables. Input time series needs to be prefixed with
    `diff_log_`.

    * `"roc"`: The gross rate of change of the LHS variables from one
    period to the next. Input time series needs to be prefixed with `roc_`.

    * `"pct"`: The percentage change of the LHS variables from one period
    to the next. Input time series needs to be prefixed with `pct_`.

???+ input "when_data"
    Specifies whether the exogenization should only occur if a valid 
    value exists in the input data.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


................................................................................
        """
        transform = _transforms.resolve_transform(transform, **kwargs, )
        self._write_to_register("exogenized", dates, names, transform, )

    @_pages.reference(category="definition_simultaneous", )
    def exogenize_anticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        *,
        status: bool | int = True,
    ) -> None:
        """
················································································

==Exogenize certain quantities at certain dates==

```
self.exogenize_anticipated(
    dates,
    names,
)
```

### Input arguments ###


???+ input "dates"

    Dates at which the `names` will be exogenized; use `...` for all simulation dates.

???+ input "names"

    Names of quantities to exogenize at the `dates`; use `...` for all exogenizable quantities.

················································································
        """
        self._write_to_register(
            "exogenized_anticipated",
            dates,
            names,
            status,
        )

    @_pages.reference(category="definition_simultaneous", )
    def exogenize_unanticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        *,
        status: bool | int = True,
    ) -> None:
        r"""
················································································

==Exogenize certain quantities at certain dates as unanticipated==

```
self.exogenize_unanticipated(
    dates, names,
    /,
    transform=None,
    when_data=False,
)
```

### Input arguments ###


???+ input "dates"

    Dates at which the `names` will be exogenized; use `...` for all simulation dates.

???+ input "names"

    Names of quantities to exogenize at the `dates`; use `...` for all exogenizable quantities.


### Input arguments available only for `Sequential` models ###

???+ input "transform"

    Transformation (specified as a string) to be applied to the exogenized
    quantities; if `None`, no tranformation is applied.

???+ input "when_data"

    If `True`, the data point will be exogenized only if a proper value
    exists in the input data.

················································································
        """
        self._write_to_register(
            "exogenized_unanticipated",
            dates,
            names,
            status,
        )

    @_pages.reference(category="information_sequential", )
    def get_exogenized_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names exogenized in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._exogenized_register,
            *args, **kwargs,
        )

    @_pages.reference(category="information_simultaneous", )
    def get_exogenized_unanticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names exogenized as unanticipated in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._exogenized_unanticipated_register,
            *args, **kwargs,
        )

    @_pages.reference(category="information_simultaneous", )
    def get_exogenized_anticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names exogenized as anticipated in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._exogenized_anticipated_register,
            *args, **kwargs,
        )

    # @_pages.reference(category="definition_sequential", )
    def endogenize(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        /,
    ) -> None:
        r"""
        """
        self._write_to_register("endogenized", dates, names, True, )

    @_pages.reference(category="definition_simultaneous", )
    def endogenize_anticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        *,
        status: bool | int = True,
    ) -> None:
        """
················································································

==Endogenize certain quantities at certain dates==

················································································
        """
        self._write_to_register(
            "endogenized_anticipated",
            dates,
            names,
            status,
        )

    @_pages.reference(category="definition_simultaneous", )
    def endogenize_unanticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        *,
        status: bool | int = True,
    ) -> None:
        """
················································································

==Endogenize certain quantities at certain dates==

················································································
        """
        self._write_to_register(
            "endogenized_unanticipated",
            dates,
            names,
            status,
        )

    @_pages.reference(category="information_simultaneous", )
    def get_endogenized_unanticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names endogenized as unanticipated in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._endogenized_unanticipated_register,
            *args,
            **kwargs,
        )

    @_pages.reference(category="information_simultaneous", )
    def get_endogenized_anticipated_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names endogenized as anticipated in a certain period==

················································································
        """
        return self._get_names_registered_in_period(
            self._endogenized_unanticipated_register,
            *args,
            **kwargs,
        )

    @property
    def any_endogenized_unanticipated_except_start(self, /, ) -> bool:
        r"""
        True if there is any endogenized unanticipated point in the plan after the first period
        """
        return self._any_in_register_except_start("endogenized_unanticipated", )

    @property
    def any_endogenized_anticipated_except_start(self, /, ) -> bool:
        r"""
        True if there is any endogenized anticipated point in the plan after the first period
        """
        return self._any_in_register_except_start("endogenized_anticipated", )

    def _any_in_register_except_start(self, register_name, ) -> bool:
        r"""
        True if there is any unanticipated point in the plan in the second
        or later simulation period
        """
        return any(
            any(_is_active_status(i) for i in v[1:])
            for v in getattr(self, f"_{register_name}_register", ).values()
        )

    @property
    def is_empty(self, /, ) -> bool:
        """
        True if there are no exogenized or endogenized points in the plan
        """
        has_any_points = any(
            _has_points_in_register(self._get_register_by_name(n, ), )
            for n in self._registers
        )
        return not has_any_points

    # def swap(
    #     self,
    #     dates: Iterable[Period] | EllipsisType,
    #     pairs: Iterable[tuple[str, str]] | tuple[str, str],
    #     *args, **kwargs,
    # ) -> None:
    #     """
    #     """
    #     pairs = tuple(pairs)
    #     if not pairs:
    #         return
    #     if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
    #         pairs = (pairs, )
    #     for pair in pairs:
    #         self.exogenize(dates, pair[0], *args, **kwargs, )
    #         self.endogenize(dates, pair[1], *args, **kwargs, )

    def swap_anticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        pairs: Iterable[tuple[str, str]] | tuple[str, str],
        *args, **kwargs,
    ) -> None:
        r"""
................................................................................

==Swap quantities as anticipated at certain dates==

Swap (exogenize and endogenize) quantities at certain dates. This method
exogenizes the first quantity in the pair and endogenizes the second
quantity in the pair at the specified dates. It is equivalent to calling
`exogenize_anticipated` and `endogenize_anticipated` separately.

    self.swap_anticipated(
        dates,
        pairs,
    )


### Input arguments ###

???+ input "self"
    The simulation plan in which data points will be exogenized and
    endogenized.

???+ input "dates"
    Dates at which the quantities will be exogenized and endogenized.

???+ input "pairs"
    A list of pairs of names to exogenize and endogenize at the specified
    dates.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


................................................................................
        """
        pairs = tuple(pairs)
        if not pairs:
            return
        if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
            pairs = (pairs, )
        for pair in pairs:
            self.exogenize_anticipated(dates, pair[0], *args, **kwargs, )
            self.endogenize_anticipated(dates, pair[1], *args, **kwargs, )

    def swap_unanticipated(
        self,
        dates: Iterable[Period] | EllipsisType,
        pairs: Iterable[tuple[str, str]] | tuple[str, str],
        *args, **kwargs,
    ) -> None:
        """
................................................................................

==Swap quantities as unanticipated at certain dates==

Swap (exogenize and endogenize) quantities at certain dates. This method
exogenizes the first quantity in the pair and endogenizes the second
quantity in the pair at the specified dates. It is equivalent to calling
`exogenize_unanticipated` and `endogenize_unanticipated` separately.

    self.swap_unanticipated(
        dates,
        pairs,
    )


### Input arguments ###

???+ input "self"
    The simulation plan in which data points will be exogenized and
    endogenized.

???+ input "dates"
    Dates at which the quantities will be exogenized and endogenized.

???+ input "pairs"
    A list of pairs of names to exogenize and endogenize at the specified
    dates.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


................................................................................
        """
        pairs = tuple(pairs)
        if not pairs:
            return
        if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
            pairs = (pairs, )
        for pair in pairs:
            self.exogenize_unanticipated(dates, pair[0], *args, **kwargs, )
            self.endogenize_unanticipated(dates, pair[1], *args, **kwargs, )

    def get_register_as_bool_array(
        self: Self,
        register_name: str,
        names: str | Iterable[str] | EllipsisType = ...,
        periods: Iterable[Period] | EllipsisType = ...,
    ) -> _np.ndarray:
        """
        """
        register = self._get_register_by_name(register_name, )
        per_indexes = self._get_per_indexes(periods, )
        names = self._resolve_validate_register_names(register, names, register_name, )
        num_names, num_pers = len(names), len(per_indexes)
        #
        if names and per_indexes:
            def get_points_for_name(name: str, ) -> tuple[bool, ...]:
                return tuple(
                    register[name][t] if t is not None else False
                    for t in per_indexes
                )
            array = _np.array(tuple(
                get_points_for_name(n, ) for n in names
            ), dtype=bool, )
        else:
            array = _np.zeros((num_names, num_pers, ), dtype=bool, )
        #
        return array

    def _get_per_indexes(
        self,
        periods: Iterable[Period] | EllipsisType,
        /,
    ) -> tuple[int | None, ...]:
        """
        """
        if periods is ...:
            return tuple(range(self.num_periods, ), )
        else:
            return tuple(
                t - self.start if self._is_per_in_span(t, ) else None
                for t in periods
            )

    def _is_per_in_span(self, per: Period, ) -> bool:
        return per >= self.start and per <= self.end

    def get_exogenized_point(
        self,
        name: str,
        date: Period,
        /,
    ) -> _transforms.PlanTransformProtocol | None:
        """
        """
        column = next((
            column
            for column, t in enumerate(self.base_span)
            if date == t
        ))
        point = self._exogenized_register[name][column]
        return point if point is not None else self._default_exogenized

    def get_endogenized_point(
        self,
        name: str,
        column: int,
        /,
    ) -> _transforms.PlanTransformProtocol | None:
        """
        """
        point = self._endogenized_register[name][column]
        return point if point is not None else self._default_endogenized

    def get_databox_names(self, /, ) -> tuple[str]:
        """
        """
        databox_names = set()
        for k, v in self._exogenized_register.items():
            databox_names.update(
                t.resolve_databox_name(k, )
                for t in v if t is not None
            )
        return tuple(n for n in databox_names if n is not None)

    def _get_names_registered_in_period(
        self,
        register: dict[str, Any],
        date: Period,
    ) -> tuple[str, ...]:
        """
        """
        column_index = self.base_span.index(date)
        return tuple(
            name
            for name, status in register.items()
            if bool(status[column_index])
        )

    def __str__(self, /, ) -> str:
        """
        """
        return self.get_pretty_string()

    def _write_to_register(
        self,
        register_name: str,
        periods: Iterable[Period] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        new_status: Any,
    ) -> None:
        """
        """
        register = self._get_register_by_name(register_name, )
        names = self._resolve_validate_register_names(register, names, register_name, )
        per_indexes, *_ = self._get_period_indexes(periods, )
        for n in names:
            for t in per_indexes:
                register[n][t] = new_status

    def _get_period_indexes(
        self,
        periods: Iterable[Period] | EllipsisType,
        /,
    ) -> tuple[tuple[int, ...], tuple[Period, ...]]:
        """
        """
        if periods is Ellipsis:
            periods = tuple(self.base_span)
        if hasattr(periods, "resolve"):
            periods = periods.resolve(self, )
        catch_invalid_periods(periods, self.base_span, )
        period_indexes = tuple(d - self.start for d in periods)
        return period_indexes, periods

    def _add_register_to_table(
        self,
        table,
        register: dict,
        action: str,
        db: Databox | None = None,
        **kwargs,
    ) -> None:
        """
        """
        def _get_status_symbol(status, ):
            return status.symbol if hasattr(status, "symbol") else _PRETTY_SYMBOL.get(status, "")
        #
        def _get_value(db, name, date, status, ):
            missing_str = Series._missing_str
            databox_name = (
                status.resolve_databox_name(name, )
                if hasattr(status, "resolve_databox_name")
                else name
            )
            try:
                value = db[databox_name][date][0, 0]
            except:
                return missing_str
            if _np.isnan(value):
                return missing_str
            return f"{value:g}"
        #
        all_rows = (
            (k, str(date), action, _get_status_symbol(status), _get_value(db, k, date, status, ), )
            for k, v in register.items()
            for status, date in zip(v, self.base_span)
            if status is not None and status is not False
        )
        #
        all_rows = sorted(all_rows, key=lambda row: (row[0], row[1], ), )
        row_groups = _it.groupby(all_rows, key=lambda row: (row[0], row[2], row[3], ), )
        for _, g in row_groups:
            representative = _create_representative_for_table_rows(tuple(g), )
            table.add_row(representative, )

    for n in _registers:
        exec(_tw.dedent(f"""
            def get_{n}(self, ) -> tuple[Period, ...]:
                return _get_registered_periods(self._{n}_register, self.base_span, )
        """))

    #]


def _get_registered_periods(
    register: dict[str, Any],
    base_span: tuple[Period],
    /,
) -> dict[str, tuple[Period]]:
    """
    """
    return {
        k: tuple(
            period
            for period, status in zip(base_span, v)
            if status is not None and status is not False
        )
        for k, v in register.items()
    }


def catch_invalid_periods(
    dates: Iterable[Period],
    base_span: tuple[Period],
    /,
) -> NoReturn | None:
    """
    """
    invalid = tuple(repr(d) for d in dates if d not in base_span)
    if invalid:
        raise _wrongdoings.IrisPieError(
            ("These date(s) are out of simulation span:", ) + invalid
        )


def _is_active_status(value: Any, /, ) -> bool:
    return (value is not None) and (value is not False)


def _has_points_in_register(register: dict, /, ) -> bool:
    return any(
        any(_is_active_status(i) for i in v)
        for v in register.values()
    )


def _create_representative_for_table_rows(rows, ):
    if len(rows) == 1:
        return rows[0]
    else:
        return (rows[0][0], rows[0][1] + ">>" + rows[-1][1], *rows[0][2:], )


Plan = SimulationPlan
PlanSimulate = SimulationPlan


_PRETTY_SYMBOL = {
    None: "",
    True: "⋅",
    False: "",
}





"""
"""


#[
from __future__ import annotations

from typing import (Self, Iterable, Literal, )
from numbers import (Real, )
import functools as _ft
import numpy as _np

from .. import wrongdoings as _wrongdoings
from .. import dates as _dates
from . import _conversions as _conversions
#]


FormType = Literal["rate", "diff", ]
AggregationType = Literal["sum", "mean", "avg", "last", "first", ]


def disaggregate_arip(
    self,
    target_period_class: _dates.Period,
    /,
    model: tuple[FormType, AggregationType],
    target: Self | None = None,
    # indicator: _np.ndarray | None = None,
) -> Self:
    """Autoregressive interpolation"""
    #[
    if not self.start_date:
        return None, None

    low_start_date = self.start_date
    low_end_date = self.end_date
    low_data_variant_iterator \
        = self.iter_own_data_variants_from_until((low_start_date, low_end_date, ), )

    low_freq = self.start_date.frequency
    num_low_periods = self.num_periods
    high_freq = target_period_class.frequency
    num_within = high_freq // low_freq
    num_high_periods = num_low_periods * num_within

    high_start_date = self.start_date.convert(high_freq, position="start", )
    high_end_date = self.end_date.convert(high_freq, position="end", )
    target_data = _get_target_data(
        target,
        high_start_date,
        high_end_date,
        num_high_periods,
    )

    high_data = disaggregate_arip_data(
        low_data_variant_iterator,
        target_data,
        model,
        num_low_periods,
        int(low_freq),
        int(high_freq),
    )

    return high_start_date, _np.column_stack(high_data, )

    #]


class _RateForm:
    """
    """
    #[

    @staticmethod
    def get_rho(low_freq, high_freq, low_data_v, ) -> Real:
        """
        """
        first_low_value, last_low_value, num_low_periods \
            = _get_first_last_observations(low_data_v, )
        low_roc = (
            ((last_low_value / first_low_value) ** (1 / num_low_periods))
            if num_low_periods else 1
        )
        return _conversions.convert_roc(low_roc, low_freq, high_freq, )

    @staticmethod
    def get_constant(*args, ) -> Real:
        return 0

    @staticmethod
    def get_sigma_vector(rho, num_high_periods, ) -> _np.ndarray:
        return rho ** _np.arange(num_high_periods, )

    #]


class _DiffForm:
    """
    """
    #[

    @staticmethod
    def get_rho(low_freq, high_freq, low_data_v, ) -> Real:
        """
        """
        return 1

    @staticmethod
    def get_constant(low_freq, high_freq, low_data_v, ) -> Real:
        """
        """
        first_low_value, last_low_value, num_low_periods \
            = _get_first_last_observations(low_data_v, )
        low_diff = (
            (last_low_value - first_low_value) / num_low_periods
            if num_low_periods else 0
        )
        return _conversions.convert_diff(low_diff, low_freq, high_freq, )

    @staticmethod
    def get_sigma_vector(rho, num_high_periods, ) -> _np.ndarray:
        return _np.ones((num_high_periods, ), dtype=float, )

    #]


def _get_first_last_observations(
    low_data_v: _np.ndarray,
) -> tuple[Real, Real]:
    """
    """
    #[
    where_finite, *_ = _np.nonzero(_np.isfinite(low_data_v, ))
    if where_finite.size:
        first_low_value = low_data_v[where_finite[0]]
        last_low_value = low_data_v[where_finite[-1]]
        num_low_periods = where_finite[-1] - where_finite[0]
    else:
        first_low_value = last_low_value = None
        num_low_periods = 0
    return first_low_value, last_low_value, num_low_periods
    #]


def _create_aggregation_vector_sum(num_within, ):
    return [1, ] * num_within


def _create_aggregation_vector_mean(num_within, ):
    return [1/num_within, ] * num_within


def _create_aggregation_vector_first(num_within, ):
    return [1, ] + [0, ]*(num_within-1)


def _create_aggregation_vector_last(num_within, ):
    return [0, ]*(num_within-1) + [1, ]


_CHOOSE_FORM = {
    "rate": _RateForm,
    "diff": _DiffForm,
}


_CHOOSE_AGGREGATION_VECTOR = {
    "sum": _create_aggregation_vector_sum,
    "mean": _create_aggregation_vector_mean,
    "avg": _create_aggregation_vector_mean,
    "first": _create_aggregation_vector_first,
    "last": _create_aggregation_vector_last,
}


CHOOSE_AGGREGATION_FUNC = {
    "sum": _np.sum,
    "mean": _np.mean,
    "avg": _np.mean,
    "first": lambda x: x[0],
    "last": lambda x: x[-1],
}


def _create_multiplier_column(low_period, num_low_periods, num_within, ):
    multiplier_column = _np.zeros((num_low_periods, num_within, ), dtype=float, )
    multiplier_column[low_period, :] = 1
    return multiplier_column.reshape(-1, 1)


def _create_target_column(high_period, num_high_periods, ):
    target_column = _np.zeros((num_high_periods, ), dtype=float, )
    target_column[high_period] = 1
    return target_column.reshape(-1, 1, )


def _create_aggregation_row(low_period, num_low_periods, num_within, aggregation_vector, ):
    out = _np.zeros((num_low_periods, num_within, ), dtype=float, )
    out[low_period, :] = aggregation_vector
    return out.reshape(1, -1, )


def _create_target_row(high_period, num_high_periods, ):
    target_row = _np.zeros((num_high_periods, ), dtype=float, )
    target_row[high_period] = 1
    return target_row.reshape(1, -1, )


def _detect_full_low_periods(target_data, num_within, ):
    x = target_data.reshape((-1, num_within, ))
    return _np.where(_np.all(_np.isfinite(x, ), axis=1, ))


def _get_target_data(
    target,
    high_start_date,
    high_end_date,
    num_high_periods,
    /,
) -> tuple[_np.ndarray, _np.ndarray, _np.ndarray]:
    """
    """
    return (
        target.get_data_from_until((high_start_date, high_end_date), )
        if target is not None else _np.full((num_high_periods, ), _np.nan, dtype=float, )
    )


def _create_basic_system_matrices(
    num_high_periods: int,
    rho: Real,
    const: Real,
    sigma_vector: _np.ndarray,
    /,
) -> tuple[_np.ndarray, _np.ndarray]:
    """
    """
    K = _np.zeros((num_high_periods-1, num_high_periods), dtype=float, )
    C = _np.full((num_high_periods-1, 1), const, dtype=float)
    for i in range(num_high_periods-1, ):
        K[i, i+1] = 1 / sigma_vector[i+1]
        K[i, i] = -rho / sigma_vector[i+1]
        C[i, 0] = const / sigma_vector[i+1]
    F = K.T @ K
    C = K.T @ C
    return F, C


def disaggregate_arip_data(
    low_data_variant_iterator: Iterable[_np.ndarray],
    target_data: _np.ndarray,
    model: tuple[str, str | tuple[Real, ...]],
    num_low_periods: int,
    low_freq: int,
    high_freq: int,
    /,
) -> tuple[_np.ndarray, ...]:
    """
    """
    where_finite_target, *_ = _np.nonzero(_np.isfinite(target_data, ))
    num_finite_target = where_finite_target.size

    num_within = high_freq // low_freq
    num_high_periods = num_low_periods * num_within
    where_full_low_periods = _detect_full_low_periods(target_data, num_within, )

    form_string, aggregation = model
    form = _CHOOSE_FORM[form_string]
    aggregation_vector = (
        _CHOOSE_AGGREGATION_VECTOR[aggregation](num_within, )
        if isinstance(aggregation, str) else tuple(aggregation)
    )

    create_multiplier_column = _ft.partial(
        _create_multiplier_column,
        num_low_periods=num_low_periods,
        num_within=num_within,
    )

    create_target_column = _ft.partial(
        _create_target_column,
        num_high_periods=num_high_periods,
    )

    create_aggregation_row = _ft.partial(
        _create_aggregation_row,
        num_low_periods=num_low_periods,
        num_within=num_within,
        aggregation_vector=aggregation_vector,
    )

    create_target_row = _ft.partial(
        _create_target_row,
        num_high_periods=num_high_periods,
    )

    high_data = ()

    for low_data_v in low_data_variant_iterator:

        if low_data_v.size != num_low_periods:
            raise _wrongdoings.IrisPieCritical(
                f"Data variant has {low_data_v.size} periods, "
                f"but {num_low_periods} are required."
            )

        if target_data.size != num_high_periods:
            raise _wrongdoings.IrisPieCritical(
                f"Target data has {target_data.size} periods, "
                f"but {num_high_periods} are required."
            )

        low_data_v[where_full_low_periods] = _np.nan
        where_finite, *_ = _np.nonzero(_np.isfinite(low_data_v, ))
        num_finite = where_finite.size

        rho = form.get_rho(low_freq, high_freq, low_data_v, )
        const = form.get_constant(low_freq, high_freq, low_data_v, )
        sigma_vector = form.get_sigma_vector(rho, num_high_periods, )

        F, C = _create_basic_system_matrices(
            num_high_periods,
            rho,
            const,
            sigma_vector,
        )

        multiplier_columns = tuple(
            create_multiplier_column(i, )
            for i in where_finite
        )

        target_columns = tuple(
            create_target_column(i, )
            for i in where_finite_target
        )

        right_padding = _np.zeros(
            (1, num_finite+num_finite_target, ),
            dtype=float,
        )

        aggregation_rows = tuple(
            _np.hstack((create_aggregation_row(i, ), right_padding, ), )
            for i in where_finite
        )

        target_rows = tuple(
            _np.hstack((create_target_row(i, ), right_padding, ), )
            for i in where_finite_target
        )

        F = _np.hstack((F, *multiplier_columns, *target_columns, ))
        F = _np.vstack((F, *aggregation_rows, *target_rows, ))

        C = _np.vstack((
            C,
            low_data_v[where_finite].reshape(-1, 1, ),
            target_data[where_finite_target].reshape(-1, 1, ),
        ))

        high_data_v = _np.linalg.solve(F, C, )
        high_data_v = high_data_v[:num_high_periods].flatten()
        high_data += (high_data_v, )

    return high_data




"""
Evaluators for dynamic period-by-period systems
"""


#[
from __future__ import annotations

import numpy as _np
from collections import (namedtuple, )

from .. import equations as _equations
from .. import quantities as _quantities
from ..aldi.maps import (ArrayMap, )
from ..incidences.main import (Token, )
from ._printers import (IterPrinter, )
from ._equators import (Equator, )
from ._jacobians import (Jacobian, )

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Any, Callable, )
    from collections.abc import (Iterable, )
#]


_Evaluator = namedtuple(
    "_Evaluator", ("evaluate", "update", "get_init_guess", "jacobian", "iter_printer", ),
)


def _create_update_map(
    wrt_tokens: Iterable[Token],
) -> ArrayMap:
    """
Create a map for updating the data array with the new guess.
    """
    update_map = ArrayMap()
    lhs_rows, lhs_columns = zip(*wrt_tokens, )
    lhs_rows = _np.array(lhs_rows, dtype=int, )
    lhs_columns = _np.array(lhs_columns, dtype=int, )
    update_map.lhs = (lhs_rows, lhs_columns, )
    return update_map

def create_evaluator(
    wrt_tokens: Iterable[Token],
    wrt_equations: Iterable[_equations.Equation],
    all_quantities: Iterable[_quantities.Quantity],
    terminal_simulator: Callable | None,
    context: dict | None,
    iter_printer_settings: dict[str, Any] | None,
    num_columns_to_eval: int,
) -> tuple[Callable, Callable]:
    """
    """
    wrt_tokens = list(wrt_tokens, )
    qid_to_logly = _quantities.create_qid_to_logly(all_quantities, )
    qid_to_name = _quantities.create_qid_to_name(all_quantities, )
    #
    # Index of loglies within wrt_tokens; needs to be list not tuple
    # because of numpy indexing
    index_logly = list(_quantities.generate_where_logly(
        ( tok.qid for tok in wrt_tokens ), qid_to_logly,
    ))

    update_map = _create_update_map(wrt_tokens, )

    equator = Equator(
        wrt_equations,
        context=context,
    )

    jacobian = Jacobian(
        wrt_equations,
        wrt_tokens,
        qid_to_logly,
        context=context,
        num_columns_to_eval=num_columns_to_eval,
    )

    iter_printer = IterPrinter(
        wrt_equations,
        tuple(tok.qid for tok in wrt_tokens),
        qid_to_logly,
        qid_to_name,
        **(iter_printer_settings or {}),
    )

    def get_init_guess(
        data_array: _np.ndarray,
        columns: int | _np.ndarray,
    ) -> _np.ndarray:
        """
        """
        column_offset = columns[0]
        maybelog_guess = data_array[update_map.lhs[0], update_map.lhs[1]+column_offset]
        maybelog_guess[index_logly] = _np.log(maybelog_guess[index_logly])
        return maybelog_guess

    def update(
        maybelog_guess: _np.ndarray,
        data_array: _np.ndarray,
        columns: int | _np.ndarray,
    ) -> None:
        """
        """
        column_offset = columns[0]
        # Create a copy of maybelog_guess to avoid modifying the original
        # array in scipy.optimize.root
        maybelog_guess = _np.copy(maybelog_guess, )
        maybelog_guess[index_logly] = _np.exp(maybelog_guess[index_logly])
        data_array[update_map.lhs[0], update_map.lhs[1]+column_offset] = maybelog_guess

    def evaluate(
        maybelog_guess: _np.ndarray | None,
        data_array: _np.ndarray,
        columns: _np.ndarray,
    ) -> tuple[_np.ndarray, _np.ndarray]:
        """
        """
        if maybelog_guess is not None:
            update(maybelog_guess, data_array, columns, )
        if terminal_simulator is not None:
            terminal_simulator(data_array, last_simulation=columns[-1], )
        #
        # The equator evaluates to a tuple of 1D arrays, each tuple element for
        # one equation across all periods. Reorganize into a 1D array ordered
        # as follows:
        #
        # [eq1[t=0], eq2[t=0], ..., eq1[t=1], eq2[t=1], ...]
        #
        equator_outcome = equator.eval(data_array, columns, )
        equator_outcome = _np.vstack(equator_outcome, ).flatten(order="Fortran"[0], )
        #
        column_offset = columns[0]
        jacobian_outcome = jacobian.eval(data_array, column_offset=columns[0], )
        #
        if maybelog_guess is not None:
            iter_printer.next(maybelog_guess, equator_outcome, jacobian_calculated=True, )
        return equator_outcome, jacobian_outcome,

    return _Evaluator(
        evaluate=evaluate,
        update=update,
        get_init_guess=get_init_guess,
        jacobian=jacobian,
        iter_printer=iter_printer,
    )

    #]





import pytest
import irispie as ir
import random as rn


rn.seed(0)


def random_series():
    return ir.Series(
        periods=ir.qq(2020,1,...,2030,4),
        func=rn.gauss,
        num_variants=5,
    )

N = 5

@pytest.mark.parametrize(
    ["index", "x"], [(i, random_series(), ) for i in range(N, ) ]
)
def test_contributions_plot(index: int, x: ir.Series, ) -> None:
    y = ir.series.sum(x)

    figure_title = f"Test chart {index}"

    # Plot bar chart with contributions, bar_relative
    info = x.plot(
        show_figure=False,
        chart_type="bar_relative",
        figure_title=figure_title,
        return_info=True,
        legend=("A", "B", "C", "D", "E", ),
    )

    f = info["figure"]

    # Plot white background for total line
    y.plot(
        figure=f,
        show_figure=False,
        update_traces={"line.width": 6, "line.color": "white"},
    )

    # Plot total line
    y.plot(
        figure=f,
        show_figure=False,
        update_traces={"line.color": "black"},
        legend=["Total"],
    )

    # Move figure title a bit down (default is 0.98 for a reason related to subplots)
    f.update_layout({"title.y": 0.95, })

    assert f.layout.title.text == figure_title
    assert f.layout.title.y == 0.95





"""
Jacobians for dynamic period-by-period systems
"""


#[
from __future__ import annotations

import numpy as _np


from ..jacobians import base
from ..aldi.maps import (ArrayMap, )

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from types import (EllipsisType, )
    from collections.abc import (Iterable, )
    from ..incidences.main import (Token, )
    from ..equations import (Equation, )
#]


def _create_map(
    eids: list[int],
    eid_to_wrt_tokens: dict[int, Any],
    tokens_in_columns_on_lhs: list[Any],
    eid_to_rhs_offset: dict[int, int],
    *,
    num_columns_to_eval: int,
    **kwargs,
) -> Self:
    """
    """
    #[
    map = ArrayMap()
    token_to_lhs_column = {
        t: i
        for i, t in enumerate(tokens_in_columns_on_lhs, )
    }
    num_eids = len(eids)
    #
    map_tuples = []
    rhs_row_offset = 0
    for eqn_enum, eid in enumerate(eids, ):
        wrt_tokens = eid_to_wrt_tokens[eid]
        for rhs_row, tok in enumerate(wrt_tokens, start=rhs_row_offset, ):
            for rhs_column in range(num_columns_to_eval, ):
                try:
                    tok_shifted = tok.shifted(rhs_column, )
                    lhs_column = token_to_lhs_column[tok_shifted]
                    lhs_row = eqn_enum + num_eids*rhs_column
                    map_tuples.append((lhs_row, lhs_column, rhs_row, rhs_column, ))
                except KeyError:
                    continue
        rhs_row_offset += len(wrt_tokens)
    #
    lhs_rows, lhs_columns, rhs_rows, rhs_columns = zip(*map_tuples, )
    map.lhs = (list(lhs_rows), list(lhs_columns), )
    # map.lhs = (list(lhs_columns), list(lhs_rows), )
    map.rhs = (list(rhs_rows), list(rhs_columns), )
    return map
    #]


class Jacobian(base.Jacobian, ):
    """
    """
    #[

    _create_map = staticmethod(_create_map)

    @staticmethod
    def _calculate_shape(
        eids: Collection[int],
        wrt_something: Collection[Any],
        num_columns_to_eval: int,
    ) -> tuple[int, int]:
        """
        """
        return len(eids)*num_columns_to_eval, len(wrt_something),

    def eval(
        self,
        data_array: _np.ndarray,
        column_offset: _np.ndarray,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(data_array, column_offset, )
        return self._create_jacobian_matrix(diff_array, )

    def _create_eid_to_wrts(
        self,
        equations: Iterable[Equation],
        all_wrt_tokens: Iterable[int],
        /,
    ) -> dict[int, tuple[int, ...]]:
        """
        """
        all_wrt_qids = set(tok.qid for tok in all_wrt_tokens)
        return {
            eqn.id: tuple(
                tok for tok in eqn.incidence
                if tok.qid in all_wrt_qids
            )
            for eqn in equations
        }

    # ===== Implement AtomFactoryProtocol =====

    def create_data_index_for_token(
        self,
        token: Token,
    ) -> tuple[int, _np.ndarray]:
        """
        """
        if not hasattr(self, "_column_indexes", ):
            self._column_indexes = _np.arange(self.num_columns_to_eval, )
        return token.qid, token.shift + self._column_indexes,

    def create_diff_for_token(
        self,
        token: Token,
        wrt_tokens: tuple[Token],
    ) -> _np.ndarray | int:
        """
        """
        num_wrts = len(wrt_tokens)
        if token is None:
            return _np.zeros((num_wrts, self.num_columns_to_eval, ), dtype=_np.float64, )
        try:
            index = wrt_tokens.index(token, )
            diff = _np.zeros((num_wrts, self.num_columns_to_eval, ), dtype=_np.float64, )
            diff[index] = 1
            return diff
        except:
            return 0

    #]




"""
Iteration printers
"""


#[
from collections.abc import (Iterable, )
from numbers import (Number, )
import numpy as _np
import scipy as _sp
import os as _os

from .. import equations as _equations
from .. import quantities as _quantities
#]


class IterPrinter:
    """
    Iterations printer for steady equator
    """
    #[

    _NAN_STRING = "•"
    _HEADER_DIVIDER_CHAR = "-"
    _MAX_LEN_EQUATION_STRING = 20
    _MAX_LEN_NAME_STRING = 10

    def __init__(
        self, 
        equations: Iterable[_equations.Equation],
        qids: Iterable[int],
        qid_to_logly: dict[int, bool],
        qid_to_name: dict[int, str],
        /,
        every: int = 5,
        header_message: str | None = None,
    ) -> None:
        """
        """
        self._populate_equations(equations, )
        self._populate_quantities(qids, qid_to_logly, qid_to_name, )
        self.reset()
        self.header_message = header_message
        self._every = every

    def reset(self, /, ) -> None:
        """
        """
        self._func_count = 0
        self._curr_f = None
        self._curr_x = None
        self._prev_x = None
        self._prev_f = None
        self.header_message = None

    @property
    def num_equations(self, /, ) -> int:
        """
        """
        return len(self._equations)

    def next(self, x, f, /, jacobian_calculated, ) -> None:
        """
        Handle next function evaluation
        """
        self._curr_f = _np.array(tuple(f))
        self._curr_x = _np.array(tuple(x))
        f_norm = _sp.linalg.norm(self._curr_f, 2)
        if self._func_count == 0:
            dimension = (self._curr_f.size, self._curr_x.size, )
            self.print_header(dimension, )
        self._last_iter_string = self.get_iter_string(f_norm, jacobian_calculated, *self.find_worst_diff_x(), *self.find_worst_equation(), )
        if self._func_count % self._every == 0:
            _print_to_width(self._last_iter_string)
        self._prev_x = self._curr_x
        self._prev_f = self._curr_f
        self._curr_f = None
        self._curr_x = None
        self._func_count += 1

    def find_worst_equation(self, /, ) -> tuple[Number, str]:
        """
        Find the largest function residual and the corresponding equation
        """
        abs_curr_f = _np.abs(self._curr_f)
        index = _np.argmax(abs_curr_f)
        worst_f = abs_curr_f[index]
        index = index % self.num_equations
        worst_equation = self._equations[index].human if self._equations is not None else ""
        worst_equation = _clip_string_exactly(worst_equation, self._MAX_LEN_EQUATION_STRING)
        return f"{worst_f:.5e}", worst_equation

    def find_worst_diff_x(self, /, ) -> tuple[str, str]:
        """
        Find the largest change in x and the corresponding quantity
        """
        if self._prev_x is None:
            return f"{self._NAN_STRING:>11}", f"{self._NAN_STRING:{self._MAX_LEN_NAME_STRING}}"
        diff_x = abs(self._curr_x - self._prev_x) if self._prev_x is not None else None
        index = _np.argmax(diff_x)
        worst_diff_x = diff_x[index]
        worst_qid = self._qids[index]
        maybelog_worst_name = self._qid_to_print[worst_qid]
        maybelog_worst_name = _clip_string_exactly(maybelog_worst_name, self._MAX_LEN_NAME_STRING)
        return f"{worst_diff_x:.5e}", maybelog_worst_name

    def print_header(self, dimension, /, ) -> None:
        """
        Print header for fuction evaluations
        """
        top_line = f"{str(self.header_message or '')}[Dimension {dimension[0]}×{dimension[1]}]"
        header = f"{'ƒ-count':>8}   {'‖ƒ‖':>11}   {'∇ƒ':>5}   {'max|∆x|':>11}   {' ':10}   {'max|ƒ|':>11}   {''}"
        len_top_line = len(top_line)
        len_header = len(header) + self._MAX_LEN_EQUATION_STRING + 1
        self._divider_line = self._HEADER_DIVIDER_CHAR * max(len_header, len_top_line, )
        upper_divider = self._divider_line
        upper_divider = self._divider_line[0:1] + top_line + self._divider_line[1 + len(top_line):]
        lower_divider = self._divider_line
        _print_to_width("", upper_divider, header, lower_divider, )

    def print_footer(self, /, ) -> None:
        """
        Print footer for iterations
        """
        _print_to_width(self._divider_line, self._last_iter_string, self._divider_line, )

    def get_iter_string(
        self,
        f_norm: Number,
        jacobian_calculated: bool,
        worst_diff_x: str | None,
        worst_diff_x_name: str | None,
        worst_f: Number,
        worst_equation: str,
        /,
    ) -> None:
        """
        Print info on current iteration
        """
        j_done_string = "√" if jacobian_calculated else "×"
        return f"{self._func_count:8g}   {f_norm:.5e}   {j_done_string:>5}   {worst_diff_x}   {worst_diff_x_name}   {worst_f}   {worst_equation}"

    #]


class BasicIterPrinter(IterPrinter, ):
    """
    """
    #[
    def _populate_equations(
        self,
        equations: Iterable[_equations.Equation],
        /,
    ) -> None:
        """
        """
        self._equations = tuple(equations)

    def _populate_quantities(
        self,
        qids: Iterable[int],
        qid_to_logly: dict[str, bool],
        qid_to_name: dict[str, str],
        /,
    ) -> None:
        self._qids = tuple(qids)
        self._qid_to_print = {
            i: _quantities.wrap_logly(qid_to_name[i], qid_to_logly.get(i, False), )
            for i in qids
        }
    #]


FlatSteadyIterPrinter = BasicIterPrinter


class NonflatSteadyIterPrinter(IterPrinter, ):
    """
    """
    #[
    def _populate_equations(
        self,
        equations: Iterable[_equations.Equation],
        /,
    ) -> None:
        """
        """
        self._equations = tuple(equations) + tuple(equations)

    def _populate_quantities(
        self,
        qids: Iterable[str],
        qid_to_logly: dict[str, bool],
        qid_to_name: dict[str, str],
        /,
    ) -> None:
        self._qids = tuple(qids) + tuple(-i for i in qids)
        self._qid_to_print = {
            i: _quantities.wrap_logly(qid_to_name[i], qid_to_logly.get(i, False), )
            for i in qids
        } | {
            -i: "∆" + _quantities.wrap_logly(qid_to_name[i], qid_to_logly.get(i, False), )
            for i in qids
        }
    #]


class _FuncNormColumn:
    pass


def _clip_string_exactly(
    full_string: str,
    max_length: int,
    /,
) -> str:
    """
    Clip string to exactly to max length
    """
    return \
        full_string[:max_length-1] + "⋯" \
        if len(full_string) > max_length else f"{full_string:{max_length}}"


def _print_to_width(*args, ) -> None:
    """
    """
    try:
        width = _os.get_terminal_size().columns
    except:
        width = None
    for text in args:
        if width is not None and len(text) > width:
            text = text[:width-1] + "⋯"
        print(text)




"""
Iteration printers for dynamic period-by-period systems
"""


#[
from __future__ import annotations

from ..evaluators import printers as _printers
#]


class IterPrinter(_printers.BasicIterPrinter, ):
    pass




"""
Simultaneous model simulation inlay
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, Protocol, runtime_checkable)
from collections.abc import (Iterable, )
import numpy as _np
import wlogging as _wl

from .. import quantities as _quantities
from .. import has_variants as _has_variants
from ..dates import (Period, )
from ..databoxes.main import (Databox, )
from ..dataslates.main import (Dataslate, )
from ..fords import solutions as _solutions
from ..fords import simulators as _ford_simulator
from ..period_by_period import simulators as _period_by_period_simulator
from ..stacked_time import simulators as _stacked_time_simulator
from ..plans.simulation_plans import (SimulationPlan, )
#]


_LOGGER = _wl.get_colored_two_liner(__name__, level=_wl.INFO, )


_SIMULATION_METHOD_DISPATCH = {
    "first_order": _ford_simulator.simulate,
    "period_by_period": _period_by_period_simulator.simulate,
    "period": _period_by_period_simulator.simulate,
    "stacked_time": _stacked_time_simulator.simulate,
    "stacked": _stacked_time_simulator.simulate,
}


_Info = dict[str, Any] | list[dict[str, Any]]


class Inlay:
    """
    """
    #[

    def simulate(
        self,
        in_db: Databox,
        span: Iterable[Period],
        /,
        *,
        plan: SimulationPlan | None = None,
        method: Literal["first_order", "period_by_period", "period", "stacked_time", "stacked", ] = "first_order",
        prepend_input: bool = True,
        target_db: Databox | None = None,
        num_variants: int | None = None,
        remove_initial: bool = True,
        remove_terminal: bool = True,
        shocks_from_data: bool = True,
        stds_from_data: bool = True,
        logging_level: int = _wl.INFO,
        #
        unpack_singleton: bool = True,
        return_info: bool = False,
        **kwargs,
    ) -> tuple[Databox, _Info]:
        """
        """
        _LOGGER.set_level(logging_level, )

        num_variants \
            = self.resolve_num_variants_in_context(num_variants, )
        _LOGGER.debug(f"Running {num_variants} variants")

        base_dates = tuple(span, )

        if plan is not None and not plan.is_empty:
            plan.check_consistency(self, base_dates, )
            extra_databox_names = plan.get_databox_names()
        else:
            plan = None
            extra_databox_names = None

        slatable = self.get_slatable_for_simulate(
            shocks_from_data=shocks_from_data,
            stds_from_data=stds_from_data,
        )

        dataslate = Dataslate.from_databox_for_slatable(
            slatable, in_db, base_dates,
            num_variants=num_variants,
            extra_databox_names=extra_databox_names,
        )

        zipped = zip(
            range(num_variants, ),
            self.iter_variants(),
            dataslate.iter_variants(),
        )

        simulation_func = _SIMULATION_METHOD_DISPATCH[method]

        #=======================================================================
        # Main loop over variants
        #
        out_info = []
        for vid, model_v, dataslate_v in zipped:
            info_v = simulation_func(
                model_v, dataslate_v, plan, vid,
                logger=_LOGGER,
                return_info=return_info,
                **kwargs,
            )
            out_info.append(info_v, )
        #=======================================================================
        #
        # Remove initial and terminal condition data (all lags and leads
        # before and after the simulation span)
        if remove_terminal:
            dataslate.remove_terminal()
        if remove_initial:
            dataslate.remove_initial()
        #
        # Convert all variants of the dataslate to a databox
        out_db = dataslate.to_databox()
        if prepend_input:
            out_db.prepend(in_db, base_dates[0]-1, )
        #
        # Add to custom databox
        if target_db is not None:
            out_db = target_db | out_db
        #
        if return_info:
            is_singleton = num_variants == 1
            out_info = _has_variants.unpack_singleton(
                out_info, is_singleton,
                unpack_singleton=unpack_singleton,
            )
            return out_db, out_info
        else:
            return out_db
    #]




"""
Dynamic nonlinear period-by-period simulator
"""


#[
from __future__ import annotations

import numpy as _np
import scipy as _sp
import functools as _ft
import wlogging as _wl

from ..simultaneous import main as _simultaneous
from ..plans.simulation_plans import (SimulationPlan, )
from ..dataslates.main import (Dataslate, )
from ..incidences.main import (Token, )
from .. import equations as _equations
from .. import equations as _equations
from .. import wrongdoings as _wrongdoings
from ..fords import simulators as _ford_simulators

from . import _evaluators as _evaluators

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from numbers import (Real, )
    from typing import (Callable, Any, Literal, )
    from ..dates import (Period, )
#]


_DEFAULT_FALLBACK_VALUE = 1/9
_METHOD_NAME = "stacked_time"


def _simulate_frame(
    simulatable_v: _simultaneous.Simultaneous,
    dataslate_v: Dataslate,
    plan: SimulationPlan | None,
    vid: int,
    *,
    logger: _wl.Logger,
    start_iter_from: Literal["data", "first_order", "previous_period"] = "previous_period",
    root_settings: dict[str, Any] | None = None,
    iter_printer_settings: dict[str, Any] | None = None,
    when_fails: Literal["critical", "error", "warning", "silent"] = "critical",
    when_missing: Literal["critical", "error", "warning", "silent"] = "critical",
    fallback_value: Real = _DEFAULT_FALLBACK_VALUE,
    terminal: Literal["data", "first_order", ] = "first_order",
    return_info: bool = False,
) -> dict[str, Any]:
    """
    """
    raise NotImplementedError()
    if not simulatable_v.is_singleton:
        raise ValueError("Simulator requires a singleton simulatable object")
    #
    root_settings = (
        _DEFAULT_ROOT_SETTINGS
        if root_settings is None
        else _DEFAULT_ROOT_SETTINGS | root_settings
    )

    qid_to_name = simulatable_v.create_qid_to_name()
    name_to_qid = simulatable_v.create_name_to_qid()
    all_quantities = simulatable_v.get_quantities()
    #
    wrt_qids = _get_wrt_qids(simulatable_v, name_to_qid, )
    wrt_equations = simulatable_v.get_dynamic_equations(kind=_equations.ENDOGENOUS_EQUATION, )

    periods = dataslate_v.periods
    base_columns = _np.array(dataslate_v.base_columns, dtype=int, )
    columns_to_run = base_columns
    num_columns_to_run = len(columns_to_run)

    if len(wrt_equations) != len(wrt_qids):
        raise _wrongdoings.IrisPieError(
            f"Number of endogenous equations {len(wrt_equations)} "
            f"does not match number of endogenous quantities {len(wrt_qids)}"
        )

    if not wrt_equations:
        return

    terminal_simulator = (
        _ford_simulators.create_terminal_simulator(simulatable_v, )
        if terminal == "first_order" else None
    )

    ford_extender = _ford_simulators.create_extender(simulatable_v, )

    create_evaluator = _ft.partial(
        _evaluators.create_evaluator,
        wrt_equations=wrt_equations,
        all_quantities=all_quantities,
        terminal_simulator=terminal_simulator,
        context=simulatable_v.get_context(),
        iter_printer_settings=iter_printer_settings,
        num_columns_to_eval=num_columns_to_run,
    )

    create_evaluator = _ft.lru_cache(maxsize=None, )(create_evaluator, )

    when_missing_stream = \
        _wrongdoings.STREAM_FACTORY[when_missing] \
        (f"These values are missing at the start of simulation: ")

    # starter = _ITER_STARTER[start_iter_from]

    catch_missing = _ft.partial(
        _catch_missing,
        qid_to_name=qid_to_name,
        fallback_value=fallback_value,
        when_missing_stream=when_missing_stream,
        periods=periods,
    )

    data = dataslate_v.get_data_variant(0, )

    current_periods = tuple(periods[i] for i in columns_to_run)
    wrt_tokens = tuple( Token(qid, shift, ) for shift in range(num_columns_to_run) for qid in wrt_qids  )
    column_offset = columns_to_run[0]

    current_wrt_qids, current_evaluator = \
        _setup_current_period(plan, create_evaluator, wrt_tokens, current_periods, name_to_qid, )
    a, b = current_evaluator.evaluate(None, data, columns_to_run, )

    current_evaluator.iter_printer.header_message = _create_header_message(vid, current_periods, )
    # data[current_wrt_qids, t] = starter(data, current_wrt_qids, t, )
    ford_extender(data, columns_to_run[0]-1, num_columns_to_run, )

    catch_missing(data, columns_to_run, )
    #
    data0 = data.copy()
    init_guess = current_evaluator.get_init_guess(data, columns_to_run, )

    root_final = _sp.optimize.root(
        current_evaluator.evaluate, init_guess,
        args=(data, columns_to_run, ),
        jac=True,
        **root_settings,
    )

    data = data0
    guess = current_evaluator.get_init_guess(data, columns_to_run, )
    for i in range(10):
        f, j = current_evaluator.evaluate(guess, data, columns_to_run, )
        #j0 = j[0,0:42]
        #index = j0.nonzero()
        #print(j0[*index])
        print(i, _np.max(_np.abs(f)))
        unit_step = - _np.linalg.solve(j, f)
        opt_step, min_f = None, None
        ff = []
        #for j in (0.1, 0.5, 1.0, 1.2, 1.5, 1.8):
        #    candidate = guess + j*unit_step
        #    f, _ = current_evaluator.evaluate(candidate, data, columns_to_run, )
        #    max_f = _np.abs(f).max()
        #    ff.append(max_f)
        #    if min_f is None or max_f < min_f:
        #        min_f = max_f
        #        opt_step = j
        opt_step = 1
        print(opt_step)
        guess = guess + opt_step*unit_step




    #
    # if not root_final.success:
        # _catch_fail(root_final, when_fails, )
    # current_evaluator.update(root_final.x, data, t, )
    # #
    # current_evaluator.iter_printer.print_footer()
    # current_evaluator.iter_printer.reset()
    #
    info = {
        "method": _METHOD_NAME,
    }
    #
    return info


def _setup_current_period(
    plan: SimulationPlan | None,
    create_evaluator: Callable,
    current_wrt_tokens: tuple[int, ...],
    current_periods: Period,
    name_to_qid: dict[str, int],
    /,
) -> tuple[tuple[int, ...], PeriodEvaluator]:
    """
    """
    current_wrt_tokens = tuple(current_wrt_tokens)
    if plan:
        names_exogenized = plan.get_exogenized_unanticipated_in_period(current_periods, )
        names_endogenized = plan.get_endogenized_unanticipated_in_period(current_periods, )
        if len(names_exogenized) != len(names_endogenized):
            raise _wrongdoings.IrisPieCritical(
                f"Number of exogenized quantities {len(names_exogenized)}"
                f" does not match number of endogenized quantities {len(names_endogenized)}"
                f" in periods {current_periods[0]}>>{current_periods[-1]}"
            )
        qids_exogenized = tuple(name_to_qid[name] for name in names_exogenized)
        qids_endogenized = tuple(name_to_qid[name] for name in names_endogenized)
        current_wrt_tokens = tuple(sorted(set(current_wrt_tokens).difference(qids_exogenized).union(qids_endogenized)))
    current_evaluator = create_evaluator(current_wrt_tokens, )
    return current_wrt_tokens, current_evaluator


def _start_iter_from_previous_period(
    data: _np.ndarray,
    wrt_qids: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    source_t = t - 1 if t > 0 else 0
    return data[wrt_qids, source_t]


def _start_iter_from_data(
    data: _np.ndarray,
    wrt_qids: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    return data[wrt_qids, t]


def _start_iter_from_first_order(
    data: _np.ndarray,
    wrt_qid: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    raise NotImplementedError()


def _catch_missing(
    data: _np.ndarray,
    columns: int,
    /,
    qid_to_name: dict[int, str],
    when_missing_stream: _wrongdoings.Stream,
    periods: Iterable[Period],
    fallback_value: Real,
) -> None:
    """
    """
    missing = _np.isnan(data[:, columns])
    if not missing.any():
        return
    data[:, columns][missing] = fallback_value
    missing_qids, *_ = missing.any(axis=1, ).nonzero()
    start = periods[columns[0]]
    end = periods[columns[-1]]
    for qid in missing_qids:
        message = f"{qid_to_name[qid]} when simulating {start}>>{end}"
        when_missing_stream.add(message, )


def _catch_fail(
    root_final: _sp.optimize.OptimizeResult,
    when_fails: Literal["error", "warning", "silent"],
) -> None:
    """
    """
    raise ValueError()


def _create_header_message(
    vid: str,
    current_periods: int,
) -> str:
    """
    """
    start = current_periods[0]
    end = current_periods[-1]
    return f"[Variant {vid}][Periods {start}>>{end}]"


def _get_wrt_qids(
    simulatable_v,
    name_to_qid: dict[str, int],
    /,
) -> tuple[int, ...]:
    """
    """
    plannable = simulatable_v.get_simulation_plannable()
    wrt_names \
        = tuple(plannable.can_be_exogenized_unanticipated) \
        + tuple(plannable.can_be_exogenized_anticipated)
    wrt_qids = sorted(
        name_to_qid[name]
        for name in set(wrt_names)
    )
    return wrt_qids


_ITER_STARTER = {
    "previous_period": _start_iter_from_previous_period,
    "data": _start_iter_from_data,
    "first_order": _start_iter_from_first_order,
}


_DEFAULT_ROOT_SETTINGS = {
    "method": "hybr",
    "tol": 1e-12,
    # "options": {"col_deriv": True},
}


simulate = _simulate_frame




"""
Descriptor for first-order systems and solutions
=================================================

Unsolved system
----------------

$$
A E[x_{t}] + B E[x_{t-1}] + C + D u_{t} + E v_{t} = 0 \\
F y_{t} + G x_{t} + H + J w_{t} = 0
$$

## State-space solution

$$
x_{t} = T x_{t-1} + K + R v_{t} \\
y_{t} = Z x_{t} + D + H w_{t}
$$


Table of contents
------------------

# List classes and standalone functions in order of appearance

* `Descriptor`
* `SystemVectors`
* `SolutionVectors`
* `HumanSolutionVectors`
* `_TRANSITION_SYSTEM_QUANTITY`
* `_MEASUREMENT_SYSTEM_QUANTITY`
* `_SYSTEM_QUANTITY`
* `_create_system_transition_vector`
* `_solution_vector_from_system_vector`
* `_get_num_forwards`
* `_get_num_backwards`
* `SystemMap`
* `_create_dynid_matrices`
* `_adjust_for_measurement_equations`
* `_custom_order_equations_by_eids`
* `Squid`

"""


#[
from __future__ import annotations

from typing import (Self, Any, Protocol, )
from collections.abc import (Iterable, )
import itertools as _it
import functools as _ft
import dataclasses as _dc
import numpy as _np

from ..incidences.main import (Token, )
from ..incidences import main as _incidence
from .. import equations as _equations
from .. import quantities as _quantities
from ..aldi import differentiators as _differentiators
from ..aldi import maps as _maps
from .. import sources as _sources
#]


class Descriptor:
    """
    Descriptor of first-order systems and solutions, and aldi Atom factory provider
    """
    #[

    __slots__ = (
        # Unsolved system vectors
        "system_vectors",

        # Solution vectors
        "solution_vectors",

        # Mapping from aldi results to unsolved matrices
        "system_map",

        # Aldi context for creating unsolved system matrices
        "aldi_context",
    )

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        quantities: Iterable[_quantities.Quantity],
        context: dict[str, Any] | None,
        /,
    ) -> None:
        """
        """
        self.system_vectors = SystemVectors(equations, quantities)
        self.solution_vectors = SolutionVectors(self.system_vectors)
        self.system_map = SystemMap(self.system_vectors)
        system_equations = _custom_order_equations_by_eids(
            equations,
            self.system_vectors.transition_eids
            + self.system_vectors.measurement_eids,
        )
        #
        # Create the evaluation context for the algorithmic differentiator
        atom_factory = self
        self.aldi_context = _differentiators.Context(
            atom_factory,
            system_equations,
            eid_to_wrts=self.system_vectors.eid_to_wrt_tokens,
            qid_to_logly=_quantities.create_qid_to_logly(quantities),
            context=context,
        )

    def get_num_backwards(self: Self) -> int:
        return self.system_vectors.get_num_backwards()

    def get_num_forwards(self: Self) -> int:
        return self.system_vectors.get_num_forwards()

    # ===== Implement AtomFactoryProtocol =====

    def create_diff_for_token(
        self,
        token: Token,
        wrt_tokens: tuple[Token, ...],
        /,
    ) -> _np.ndarray:
        """
        """
        if token is None:
            return _np.zeros((len(wrt_tokens), 1, ), )
        try:
            index = wrt_tokens.index(token, )
            diff = _np.zeros((len(wrt_tokens), 1, ), )
            diff[index] = 1
            return diff
        except:
            return 0

    def create_data_index_for_token(
        self,
        token: Token,
    ) -> tuple[int, slice]:
        """
        """
        return (token.qid, token.shift, )

    #]


"""
Quantity kinds that can be used in the transition part of a first-order system
"""
_TRANSITION_SYSTEM_QUANTITY = (
    _quantities.QuantityKind.TRANSITION_VARIABLE
    | _quantities.QuantityKind.UNANTICIPATED_SHOCK
    | _quantities.QuantityKind.ANTICIPATED_SHOCK
)


"""
Quantity kinds that can be used in the measurement part of a first-order system
"""
_MEASUREMENT_SYSTEM_QUANTITY = (
    _quantities.QuantityKind.MEASUREMENT_VARIABLE
    | _quantities.QuantityKind.TRANSITION_VARIABLE
    | _quantities.QuantityKind.MEASUREMENT_SHOCK
)


"""
Quantity kinds that can be used in a first-order system
"""
_SYSTEM_QUANTITY = (
    _TRANSITION_SYSTEM_QUANTITY
    | _MEASUREMENT_SYSTEM_QUANTITY
)


@_dc.dataclass(slots=True, )
class SystemVectors:
    """
    Vectors of quantities and equation ids in first-order system matrices
    """
    #[
    transition_eids: Iterable[int] | None = None
    measurement_eids: Iterable[int] | None = None
    eid_to_wrt_tokens: dict[int, Iterable[Token]] | None = None
    #
    transition_variables: Iterable[Token] | None = None
    transition_variables_are_logly: list[bool] | None = None
    are_initial_conditions: list[bool] | None = None,
    unanticipated_shocks: tuple[Token, ...] | None = None
    anticipated_shocks: tuple[Token, ...] | None = None
    measurement_variables: Iterable[Token] | None = None
    measurement_variables_are_logly: list[bool] | None = None
    measurement_shocks: tuple[Token, ...] | None = None
    #
    shape_A_excl_dynid: tuple[int, int] | None = None
    shape_B_excl_dynid: tuple[int, int] | None = None
    shape_C_excl_dynid: tuple[int] | None = None
    shape_D_excl_dynid: tuple[int, int] | None = None
    shape_E_excl_dynid: tuple[int, int] | None = None
    shape_F: tuple[int, int] | None = None
    shape_G: tuple[int, int] | None = None
    shape_H: tuple[int] | None = None
    shape_J: tuple[int, int] | None = None

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        quantities: Iterable[_quantities.Quantity],
        /,
    ) -> None:
        """
        Construct system vectors from a list of equations and a list of quantities
        """
        qid_to_logly = _quantities.create_qid_to_logly(quantities, )
        #
        self.transition_eids \
            = sorted([eqn.id for eqn in equations if eqn.kind in _equations.EquationKind.TRANSITION_EQUATION])
        self.measurement_eids \
            = sorted([eqn.id for eqn in equations if eqn.kind in _equations.EquationKind.MEASUREMENT_EQUATION])
        qid_to_kind = _quantities.create_qid_to_kind(quantities, )
        #
        # Collect all tokens from equations but also add zero-shifted tokens
        # created from all variables, shocks, and parameters to catch any names
        # not used in equations (shocks, parameters)
        all_tokens = _collect_all_tokens(equations, quantities, )
        #
        all_wrt_tokens = set(_incidence.generate_tokens_of_kinds(all_tokens, qid_to_kind, _SYSTEM_QUANTITY))
        self.eid_to_wrt_tokens = _equations.create_eid_to_wrt_tokens(equations, all_wrt_tokens)
        #
        actual_tokens_transition_variables \
            = set(_incidence.generate_tokens_of_kinds(all_tokens, qid_to_kind, _quantities.QuantityKind.TRANSITION_VARIABLE))
        #
        # Make adjustment for transition variables in measurement
        # equations: each x(t-k) in measurement needs to be in the current
        # dated (LHS) vector of transition variables; this is done by
        # pretending x(t-k-1) is needed
        #
        adjusted_tokens_transition_variables \
            = _adjust_for_measurement_equations(actual_tokens_transition_variables, equations, qid_to_kind)
        #
        self.transition_variables \
            = _incidence.sort_tokens(_create_system_transition_vector(adjusted_tokens_transition_variables))
        self.transition_variables_are_logly \
            = [ qid_to_logly[tok.qid] for tok in self.transition_variables ]
        self.are_initial_conditions = [
            Token(t.qid, t.shift-1) in actual_tokens_transition_variables and t.shift <= 0
            for t in self.transition_variables
        ]
        #
        # Unanticipated shocks
        #
        self.unanticipated_shocks \
            = tuple(_incidence.sort_tokens(_incidence.generate_tokens_of_kinds(
                all_tokens, qid_to_kind, _quantities.QuantityKind.UNANTICIPATED_SHOCK,
            )))
        #
        # Anticipated shocks
        #
        self.anticipated_shocks \
            = tuple(_incidence.sort_tokens(_incidence.generate_tokens_of_kinds(
                all_tokens, qid_to_kind, _quantities.QuantityKind.ANTICIPATED_SHOCK,
            )))
        #
        # Measurement variables
        #
        self.measurement_variables \
            = tuple(_incidence.sort_tokens(_incidence.generate_tokens_of_kinds(
                all_tokens, qid_to_kind, _quantities.QuantityKind.MEASUREMENT_VARIABLE,
            )))
        self.measurement_variables_are_logly \
            = [ qid_to_logly[tok.qid] for tok in self.measurement_variables ]
        #
        # Measurement shocks
        #
        self.measurement_shocks \
            = tuple(_incidence.sort_tokens(_incidence.generate_tokens_of_kinds(
                all_tokens, qid_to_kind, _quantities.QuantityKind.MEASUREMENT_SHOCK,
            )))
        #
        # Shapes of matrices
        #
        self.shape_A_excl_dynid = (len(self.transition_eids), len(self.transition_variables), )
        self.shape_B_excl_dynid = self.shape_A_excl_dynid
        self.shape_C_excl_dynid = (len(self.transition_eids), )
        self.shape_D_excl_dynid = (len(self.transition_eids), len(self.unanticipated_shocks), )
        self.shape_E_excl_dynid = (len(self.transition_eids), len(self.anticipated_shocks), )
        self.shape_F = (len(self.measurement_eids), len(self.measurement_variables), )
        self.shape_G = (len(self.measurement_eids), len(self.transition_variables), )
        self.shape_H = (len(self.measurement_eids), )
        self.shape_J = (len(self.measurement_eids), len(self.measurement_shocks), )

    def get_num_backwards(self) -> int:
        return _get_num_backwards(self.transition_variables)

    def get_num_forwards(self) -> int:
        return _get_num_forwards(self.transition_variables)
    #]


@_dc.dataclass(slots=True, )
class SolutionVectors:
    """
    Vectors of quantities in first-order solution matrices
    """
    #[
    transition_variables: tuple[Token, ...] | None = None
    are_initial_conditions: list[bool, ...] | None = None
    unanticipated_shocks: tuple[Token, ...] | None = None
    anticipated_shocks: tuple[Token, ...] | None = None
    measurement_variables: tuple[Token, ...] | None = None
    measurement_shocks: tuple[Token, ...] | None = None

    def __init__(self, system_vectors: SystemVectors, /, ) -> None:
        """
        Construct solution vectors and initial conditions indicator
        """
        self.transition_variables, self.are_initial_conditions = \
            _solution_vector_from_system_vector(system_vectors.transition_variables, system_vectors.are_initial_conditions)
        self.unanticipated_shocks = tuple(system_vectors.unanticipated_shocks)
        self.anticipated_shocks = tuple(system_vectors.anticipated_shocks)
        self.measurement_variables = tuple(system_vectors.measurement_variables)
        self.measurement_shocks = tuple(system_vectors.measurement_shocks)

    def get_initials(
        self,
        /,
    ) -> list[Token]:
        """
        Get tokens representing required initial conditions
        """
        return list(_it.compress(self.transition_variables, self.are_initial_conditions))

    def get_curr_transition_indexes(
        self,
        /,
    ) -> tuple[tuple[int, ...], tuple[int, ...]]:
        """
        """
        lhs_rhs_tuples = [
            (t.qid, i)
            for i, t in enumerate(self.transition_variables, )
            if not t.shift
        ]
        return tuple(zip(*lhs_rhs_tuples))

    #]


@_dc.dataclass(slots=True, )
class HumanSolutionVectors:
    """
    """
    #[

    transition_variables: tuple[str, ...] | None = None
    unanticipated_shocks: tuple[str, ...] | None = None
    anticipated_shocks: tuple[str, ...] | None = None
    measurement_variables: tuple[str, ...] | None = None
    measurement_shocks: tuple[str, ...] | None = None

    def __init__(
        self,
        solution_vectors: SolutionVectors,
        qid_to_name: dict[int, str],
        qid_to_logly: dict[int, bool],
        /,
    ) -> None:
        """
        """
        print_tokens = _ft.partial(_incidence.print_tokens, qid_to_name=qid_to_name, qid_to_logly=qid_to_logly, )
        for a in self.__slots__:
            setattr(self, a, print_tokens(getattr(solution_vectors, a)))

    #]


def _create_system_transition_vector(
    tokens_transition_variables: Iterable[Token],
    /,
) -> Iterable[Token]:
    """
    From tokens of transition variables, create vector of transition variables
    along columns of matrix A in unsolved system
    """
    #[
    tokens_transition_variables = set(tokens_transition_variables)
    min_shifts = _incidence.get_some_shift_by_quantities(tokens_transition_variables, lambda x: min(min(x), -1))
    max_shifts = _incidence.get_some_shift_by_quantities(tokens_transition_variables, max)
    #
    vector_for_id = lambda qid: [Token(qid, sh) for sh in range(min_shifts[qid]+1, max_shifts[qid]+1)]
    unique_ids = set(t.qid for t in tokens_transition_variables)
    return _it.chain.from_iterable(vector_for_id(i) for i in unique_ids)
    #]


def _solution_vector_from_system_vector(
    system_transition_vector: Iterable[Token],
    are_initial_conditions: Iterable[bool],
    /,
) -> Iterable[Token]:
    """
    From sorted system vector, get vector of transition variables in solved
    system and the indicator of required initial conditions
    """
    num_forwards = _get_num_forwards(system_transition_vector)
    return (
        tuple(system_transition_vector[num_forwards:]),
        tuple(are_initial_conditions[num_forwards:]),
    )


def _get_num_forwards(system_transition_vector: Iterable[Token]):
    """
    Number of forward-looking tokens in a vector of tokens
    """
    return sum(1 for t in system_transition_vector if t.shift>0)


def _get_num_backwards(system_transition_vector: Iterable[Token]):
    """
    Number of backward-looking tokens in a vector of tokens
    """
    return len(system_transition_vector) - _get_num_forwards(system_transition_vector)


class SystemMap:
    """
    """
    #[

    __slots__ = (
        "A",
        "B",
        "C",
        "D",
        "E",
        "dynid_A",
        "dynid_B",
        "dynid_C",
        "dynid_D",
        "dynid_E",
        "F",
        "G",
        "H",
        "J",
    )

    def __init__(
        self,
        system_vectors: SystemVectors,
    ) -> None:
        """
        """
        system_eids = system_vectors.transition_eids + system_vectors.measurement_eids
        #
        # Create the map from equation ids to rhs offset; the offset is the
        # number of rows in the Jacobian matrix that precede the equation
        eid_to_rhs_offset = _maps.create_eid_to_rhs_offset(system_eids, system_vectors.eid_to_wrt_tokens)
        #
        # Transition equations
        #
        self.A = _maps.ArrayMap.static(
            system_vectors.transition_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.transition_variables,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        lagged_transition_variables = [
            t.shifted(-1)
            for t in system_vectors.transition_variables
        ]
        #
        lagged_transition_variables = [
            t if t not in system_vectors.transition_variables else None
            for t in lagged_transition_variables
        ]
        #
        self.B = _maps.ArrayMap.static(
            system_vectors.transition_eids,
            system_vectors.eid_to_wrt_tokens,
            lagged_transition_variables,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self.A.remove_nones()
        self.B.remove_nones()
        #
        self.C = _maps.VectorMap.static(system_vectors.transition_eids, )
        #
        self.D = _maps.ArrayMap.static(
            system_vectors.transition_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.unanticipated_shocks,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self.E = _maps.ArrayMap.static(
            system_vectors.transition_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.anticipated_shocks,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        num_dynid_rows = len(system_vectors.transition_variables) - len(system_vectors.transition_eids)
        self.dynid_A, self.dynid_B = _create_dynid_matrices(system_vectors.transition_variables, )
        self.dynid_C = _np.zeros((num_dynid_rows, ), dtype=float, )
        self.dynid_D = _np.zeros((num_dynid_rows, system_vectors.shape_D_excl_dynid[1]), dtype=float, )
        self.dynid_E = _np.zeros((num_dynid_rows, system_vectors.shape_E_excl_dynid[1]), dtype=float, )
        #
        # Measurement equations
        #
        self.F = _maps.ArrayMap.static(
            system_vectors.measurement_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.measurement_variables,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self.G = _maps.ArrayMap.static(
            system_vectors.measurement_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.transition_variables,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self.H = _maps.VectorMap.static(system_vectors.measurement_eids, )
        #
        self.J = _maps.ArrayMap.static(
            system_vectors.measurement_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.measurement_shocks,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
    #]


def _create_dynid_matrices(system_transition_vector: Iterable[Token]):
    """
    Create dynamic identity matrix for unsolved system
    """
    #[
    max_shifts = _incidence.get_some_shift_by_quantities(system_transition_vector, max)
    #
    index_A = ([], [])
    index_B = ([], [])
    row_count = 0
    for i, t in enumerate(system_transition_vector):
        if t.shift==max_shifts[t.qid]:
            continue
        j = system_transition_vector.index(t.shifted(+1))
        index_A[0].append(row_count)
        index_A[1].append(i)
        index_B[0].append(row_count)
        index_B[1].append(j)
        row_count += 1
    #
    num_columns = len(system_transition_vector)
    dynid_A = _np.zeros((row_count, num_columns), dtype=float)
    dynid_B = _np.zeros((row_count, num_columns), dtype=float)
    dynid_A[index_A] = 1
    dynid_B[index_B] = -1
    return dynid_A, dynid_B
    #]


def _adjust_for_measurement_equations(
    tokens_transition_variables: Iterable[_quantities.Quantity],
    equations: Iterable[_equations.Equation],
    qid_to_kind: dict[int, _quantities.QuantityKind],
    /,
) -> Iterable[_quantities.Quantity]:
    """
    """
    #[
    tokens_in_measurement_equations = _it.chain.from_iterable(e.incidence for e in equations if e.kind is _equations.EquationKind.MEASUREMENT_EQUATION)
    pretend_needed = [
        Token(t.qid, t.shift-1) for t in tokens_in_measurement_equations
        if qid_to_kind[t.qid] in _quantities.QuantityKind.TRANSITION_VARIABLE
    ]
    return set(tokens_transition_variables).union(pretend_needed)
    #]


def _custom_order_equations_by_eids(
    equations: Iterable[_equations.Equation],
    eids: list[int],
    /,
) -> tuple[_equations.Equation, ...]:
    """
    """
    #[
    eid_to_equation = { eqn.id: eqn for eqn in equations }
    return tuple( eid_to_equation[eid] for eid in eids )
    #]


class SquidableProtocol(Protocol, ):
    """
    """
    #[

    solution_vectors: SolutionVectors
    shock_qid_to_std_qid: dict[int, int]

    #]


class Squid:
    """
    """
    #[

    __slots__ = (
        "curr_xi_qids",
        "curr_xi_indexes",
        "y_qids",
        "u_qids",
        "v_qids",
        "w_qids",
        "std_u_qids",
        "std_v_qids",
        "std_w_qids",
        "num_xi",
    )

    def __init__(
        self,
        model: SquidableProtocol,
        /,
    ) -> None:
        """
        """
        vec = model.solution_vectors
        shock_qid_to_std_qid = model.shock_qid_to_std_qid
        #
        self.num_xi = len(vec.transition_variables)
        self.curr_xi_qids, self.curr_xi_indexes = vec.get_curr_transition_indexes()
        self.curr_xi_qids = tuple(self.curr_xi_qids)
        self.curr_xi_indexes = tuple(self.curr_xi_indexes)
        #
        self.y_qids = tuple(t.qid for t in vec.measurement_variables)
        self.u_qids = tuple(t.qid for t in vec.unanticipated_shocks)
        self.v_qids = tuple(t.qid for t in vec.anticipated_shocks)
        self.w_qids = tuple(t.qid for t in vec.measurement_shocks)
        #
        self.std_u_qids = None
        self.std_v_qids = None
        self.std_w_qids = None
        #
        if not model.is_deterministic:
            self.std_u_qids = tuple(
                shock_qid_to_std_qid[t.qid]
                for t in vec.unanticipated_shocks
            )
            self.std_v_qids = tuple(
                shock_qid_to_std_qid[t.qid]
                for t in vec.anticipated_shocks
            )
            self.std_w_qids = tuple(
                shock_qid_to_std_qid[t.qid]
                for t in vec.measurement_shocks
            )

    @property
    def num_curr_xi(self) -> int:
        """==Number of current-dated transition variables in transition vector=="""
        return len(self.curr_xi_qids)

    @property
    def num_u(self) -> int:
        """==Number of unanticipated shocks in system=="""
        return len(self.u_qids)

    @property
    def num_v(self) -> int:
        """==Number of anticipated shocks in system=="""
        return len(self.v_qids)

    @property
    def num_w(self) -> int:
        """==Number of measurement shocks in system=="""
        return len(self.w_qids)

    #]


def _collect_all_tokens(
    equations: Iterable[_equations.Equation],
    quantities: Iterable[_quantities.Quantity],
    /,
) -> set[Token]:
    """
    Combine all tokens from equations and zero-shifted tokens created from
    variables, shocks and parameters.
    """
    #[
    # Tokens from equations
    tokens_from_equations \
        = set(_equations.generate_all_tokens_from_equations(equations, ))
    #
    # Zero-shift tokens from all variables, shocks and parameters
    kind = _quantities.ANY_VARIABLE | _quantities.ANY_SHOCK | _quantities.PARAMETER
    tokens_from_quantities \
        = set(_incidence.generate_zero_shift_tokens_from_quantities(quantities, kind=kind))
    return tokens_from_equations | tokens_from_quantities
    #]




"""
"""


#[
from __future__ import annotations

from prettytable import (PrettyTable, )
import functools as _ft

from .. import quantities as _quantities
from .. import file_io as _file_io

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Any, )
    from collections.abc import (Iterable, )
#]


_DEFAULT_ROUND_TO = 16


class _PrettyTable(PrettyTable, ):
    """
    """
    #[

    def to_csv(
        self,
        filename: str,
    ) -> None:
        """
        """
        _file_io.save_text(filename, self.get_csv_string(), )

    #]


class Inlay:
    """
    """
    #[

    def create_steady_table(
        self,
        columns: Iterable[str],
        kind: int = _quantities.ANY_VARIABLE | _quantities.PARAMETER,
        save_to_csv: str | None = None,
        **kwargs,
    ) -> _PrettyTable:
        """
        """
        column_constructors = [
            _COLUMN_CONSTRUCTORS[column.lower().strip()]
            for column in columns
        ]
        row_names = self.get_names(kind=kind, )
        table = _PrettyTable()
        for constructor in column_constructors:
            for header, values, settings in constructor(self, row_names, **kwargs, ):
                table.add_column(header, values, **settings, )
        if save_to_csv:
            table.to_csv(save_to_csv, )
        return table

    #]


def _name_column(
    self,
    row_names: tuple[str, ...],
    **kwargs,
) -> Iterable[str, tuple[str], dict[str, Any]]:
    """
    """
    yield "NAME", row_names, {"align": "l"}


def _empty_column(
    self,
    row_names: tuple[str, ...],
    **kwargs,
) -> Iterable[str, tuple[str], dict[str, Any]]:
    """
    """
    yield "", ("", ) * len(row_names), {}


def _description_column(
    self,
    row_names: tuple[str, ...],
    **kwargs,
) -> Iterable[str, tuple[str], dict[str, Any]]:
    """
    """
    #[
    name_to_description = _quantities.create_name_to_description(self._invariant.quantities, )
    descriptions = tuple(
        name_to_description.get(name, "")
        for name in row_names
    )
    yield "DESCRIPTION", descriptions, {"align": "l"}
    #]


def _kind_column(
    self,
    row_names: tuple[str, ...],
    **kwargs,
) -> Iterable[str, tuple[str], dict[str, Any]]:
    """
    """
    #[
    name_to_kind = _quantities.create_name_to_kind(self._invariant.quantities, )
    kinds = tuple(
        str(name_to_kind.get(name, "")).split(".", )[-1].upper()
        for name in row_names
    )
    yield "KIND", kinds, {"align": "l"}
    #]


def _comparison_column(
    self,
    row_names: tuple[str, ...],
    **kwargs,
) -> Iterable[str, tuple[str], dict[str, Any]]:
    """
    """
    #[
    name_to_logly = _quantities.create_name_to_logly(self._invariant.quantities, )
    comparisons = tuple(
        "ratio/" if name_to_logly.get(name, False) else "diff-"
        for name in row_names
    )
    yield "COMPARISON", comparisons, {"align": "c"}
    #]


def _steady_level_column(
    self,
    row_names: tuple[str, ...],
    round_to: int = _DEFAULT_ROUND_TO,
    **kwargs,
) -> Iterable[str, tuple[str], dict[str, Any]]:
    """
    """
    #[
    def _display_value(value: Real, ):
        return value if round_to is None else round(value, round_to, )
    iter_levels = (self.get_steady_levels() | self.get_parameters()).iter_variants()
    num_variants = self.num_variants
    header = "STEADY_LEVEL_{}" if num_variants > 1 else "STEADY_LEVEL"
    for i, levels in zip(range(num_variants), iter_levels):
        yield (
            header.format(i, ),
            tuple( _display_value(levels[name]) for name in row_names ),
            {"align": "r"},
        )
    #]


def _compare_steady_value(
    orig_value_iterator,
    self,
    row_names,
    round_to: int = _DEFAULT_ROUND_TO,
    **kwargs,
) -> Iterable[str, tuple[str], dict[str, Any]]:
    """
    """
    #[
    name_to_logly = _quantities.create_name_to_logly(self._invariant.quantities, )
    _, base_values, *_ = next(orig_value_iterator(self, row_names, **kwargs, ), )
    comparison_func = {
        name: (lambda x, y: x/y) if name_to_logly.get(name, False) else (lambda x, y: x-y)
        for name in row_names
    }
    def _compare(values: Iterable[Real], ) -> Iterable[Real]:
        return tuple(
            comparison_func[name](value, base_value, ) if value is not "" else ""
            for name, value, base_value in zip(row_names, values, base_values, )
        )
    for header, values, settings in orig_value_iterator(self, row_names, **kwargs, ):
        yield "COMPARE_" + header, _compare(values, ), settings
    #]


def _steady_change_column(
    self,
    row_names: tuple[str, ...],
    round_to: int = _DEFAULT_ROUND_TO,
    **kwargs,
) -> Iterable[str, tuple[str], dict[str, Any]]:
    """
    """
    #[
    def _display_value(value: Real, ):
        return (
            value if (round_to is None or value == "")
            else round(value, round_to, )
        )
    iter_changes = (self.get_steady_changes()).iter_variants()
    num_variants = self.num_variants
    header = "STEADY_CHANGE_{}" if num_variants > 1 else "STEADY_CHANGE"
    for i, changes in zip(range(num_variants), iter_changes):
        yield (
            header.format(i, ),
            tuple( _display_value(changes.get(name, ""), ) for name in row_names ),
            {"align": "r"},
        )
    #]


_COLUMN_CONSTRUCTORS = {
    "name": _name_column,
    "empty": _empty_column,
    "description": _description_column,
    "kind": _kind_column,
    "comparison": _comparison_column,
    "steady_level": _steady_level_column,
    "steady_change": _steady_change_column,
    "compare_steady_level": _ft.partial(_compare_steady_value, _steady_level_column, ),
    "compare_steady_change": _ft.partial(_compare_steady_value, _steady_change_column, ),
}


